import numpy as np
import pandas as pd
from plotly.subplots import make_subplots

from . import checks, plot, utils
from ... import _theme, _global_checks
from ..._explanation import Explanation


class Shap(Explanation):
    """Calculate predict-level variable attributions as Shapley Values

    Parameters
    -----------
    path : list of int, optional
        If specified, then attributions for this path will be plotted
        (default is `'average'`, which plots attribution means for `B` random paths).
    B : int, optional
        Number of random paths to calculate variable attributions (default is `25`).
    keep_distributions : bool, optional
        Save the distribution of partial predictions (default is `False`).
    processes : int, optional
        Number of parallel processes to use in calculations. Iterated over `B`
        (default is `1`, which means no parallel computation).
    random_state : int, optional
        Set seed for random number generator (default is random seed).

    Attributes
    -----------
    result : pd.DataFrame
        Main result attribute of an explanation.
    prediction : float
        Prediction for `new_observation`.
    intercept : float
        Average prediction for `data`.
    path : list of int or 'average'
        Path for which the attributions will be plotted.
    B : int
        Number of random paths to calculate variable attributions.
    keep_distributions : bool
        Save the distribution of partial predictions.
    yhats_distributions : pd.DataFrame or None
        The distribution of partial predictions.
    processes : int
        Number of parallel processes to use in calculations. Iterated over `B`.
    random_state : int or None
        Seed that was set for random number generator.

    Notes
    --------
    - https://pbiecek.github.io/ema/shapley.html
    """

    def __init__(self,
                 path="average",
                 B=25,
                 keep_distributions=False,
                 processes=1,
                 random_state=None):

        _path = checks.check_path(path)
        _processess = checks.check_processes(processes)
        _random_state = checks.check_random_state(random_state)

        self.path = _path
        self.keep_distributions = keep_distributions
        self.B = B
        self.result = pd.DataFrame()
        self.yhats_distributions = None
        self.prediction = None
        self.intercept = None
        self.processes = _processess
        self.random_state = _random_state

    def _repr_html_(self):
        return self.result._repr_html_()

    def fit(self,
            explainer,
            new_observation):
        """Calculate the result of explanation

        Fit method makes calculations in place and changes the attributes.

        Parameters
        -----------
        explainer : Explainer object
            Model wrapper created using the Explainer class.
        new_observation : pd.Series or np.ndarray
            An observation for which a prediction needs to be explained.

        Returns
        -----------
        None
        """

        _new_observation = checks.check_new_observation(new_observation, explainer)
        checks.check_columns_in_new_observation(_new_observation, explainer)
        self.result, self.prediction, self.intercept, self.yhats_distributions = utils.shap(
            explainer,
            _new_observation,
            self.path,
            self.keep_distributions,
            self.B,
            self.processes,
            self.random_state
        )

    def plot(self,
             objects=None,
             baseline=None,
             max_vars=10,
             digits=3,
             rounding_function=np.around,
             bar_width=16,
             min_max=None,
             vcolors=None,
             title="Shapley Values",
             vertical_spacing=None,
             show=True):
        """Plot the Shapley Values explanation

        Parameters
        -----------
        objects : Shap object or array_like of Shap objects
            Additional objects to plot in subplots (default is `None`).
        baseline: float, optional
            Starting x point for bars (default is average prediction).
        max_vars : int, optional
            Maximum number of variables that will be presented for for each subplot
            (default is `10`).
        digits : int, optional
            Number of decimal places (`np.around`) to round contributions.
            See `rounding_function` parameter (default is `3`).
        rounding_function : function, optional
            A function that will be used for rounding numbers (default is `np.around`).
        bar_width : float, optional
            Width of bars in px (default is `16`).
        min_max : 2-tuple of float, optional
            Range of OX axis (default is `[min-0.15*(max-min), max+0.15*(max-min)]`).
        vcolors : 3-tuple of str, optional
            Color of bars (default is `["#8bdcbe", "#f05a71"]`).
        title : str, optional
            Title of the plot (default is `"Shapley Values"`).
        vertical_spacing : float <0, 1>, optional
            Ratio of vertical space between the plots (default is `0.2/number of rows`).
        show : bool, optional
            `True` shows the plot; `False` returns the plotly Figure object that can 
            be edited or saved using the `write_image()` method (default is `True`).

        Returns
        -----------
        None or plotly.graph_objects.Figure
            Return figure that can be edited or saved. See `show` parameter.
        """

        # are there any other objects to plot?
        if objects is None:
            n = 1
            _result_list = [self.result.loc[self.result['B'] == 0,].copy()]
            _intercept_list = [self.intercept]
            _prediction_list = [self.prediction]
        elif isinstance(objects, self.__class__):  # allow for objects to be a single element
            n = 2
            _result_list = [self.result.loc[self.result['B'] == 0,].copy(),
                            objects.result.loc[objects.result['B'] == 0,].copy()]
            _intercept_list = [self.intercept, objects.intercept]
            _prediction_list = [self.prediction, objects.prediction]
        elif isinstance(objects, (list, tuple)):  # objects as tuple or array
            n = len(objects) + 1
            _result_list = [self.result.loc[self.result['B'] == 0,].copy()]
            _intercept_list = [self.intercept]
            _prediction_list = [self.prediction]
            for ob in objects:
                _global_checks.global_check_object_class(ob, self.__class__)
                _result_list += [ob.result.loc[ob.result['B'] == 0,].copy()]
                _intercept_list += [ob.intercept]
                _prediction_list += [ob.prediction]
        else:
            _global_checks.global_raise_objects_class(objects, self.__class__)

        # TODO: add intercept and prediction list update for multi-class
        # deleted_indexes = []
        # for i in range(n):
        #     result = _result_list[i]
        #
        #     if len(result['label'].unique()) > 1:
        #         n += len(result['label'].unique()) - 1
        #         # add new data frames to list
        #         _result_list += [v for k, v in result.groupby('label', sort=False)]
        #         deleted_indexes += [i]
        #
        # _result_list = [j for i, j in enumerate(_result_list) if i not in deleted_indexes]
        model_names = [result.iloc[0, result.columns.get_loc("label")] for result in _result_list]

        if vertical_spacing is None:
            vertical_spacing = 0.2 / n

        fig = make_subplots(rows=n, cols=1,
                            shared_xaxes=True, vertical_spacing=vertical_spacing,
                            x_title='contribution', subplot_titles=model_names)
        plot_height = 78 + 71

        if vcolors is None:
            vcolors = _theme.get_break_down_colors()

        if min_max is None:
            temp_min_max = [np.Inf, -np.Inf]
        else:
            temp_min_max = min_max

        for i, _result in enumerate(_result_list):
            if _result.shape[0] <= max_vars:
                m = _result.shape[0]
            else:
                m = max_vars + 1

            if baseline is None:
                baseline = _intercept_list[i]
            prediction = _prediction_list[i]

            df = plot.prepare_data_for_shap_plot(_result, baseline, prediction, max_vars, rounding_function, digits)

            fig.add_shape(
                type='line',
                x0=baseline,
                x1=baseline,
                y0=-1,
                y1=m,
                yref="paper",
                xref="x",
                line={'color': "#371ea3", 'width': 1.5, 'dash': 'dot'},
                row=i + 1, col=1
            )

            fig.add_bar(
                orientation="h",
                y=df['variable'].tolist(),
                x=df['contribution'].tolist(),
                textposition="outside",
                text=df['label_text'].tolist(),
                marker_color=[vcolors[int(c)] for c in df['sign'].tolist()],
                base=baseline,
                hovertext=df['tooltip_text'].tolist(),
                hoverinfo='text',
                hoverlabel={'bgcolor': 'rgba(0,0,0,0.8)'},
                showlegend=False,
                row=i + 1, col=1
            )

            fig.update_yaxes({'type': 'category', 'autorange': 'reversed', 'gridwidth': 2, 'automargin': True,
                              'ticks': 'outside', 'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True},
                             row=i + 1, col=1)

            fig.update_xaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True,
                              'ticks': "outside", 'tickcolor': 'white', 'ticklen': 3, 'fixedrange': True},
                             row=i + 1, col=1)

            plot_height += m * bar_width + (m + 1) * bar_width / 4

            if min_max is None:
                cum = df.contribution.values + baseline
                min_max_margin = cum.ptp() * 0.15
                temp_min_max[0] = np.min([temp_min_max[0], cum.min() - min_max_margin])
                temp_min_max[1] = np.max([temp_min_max[1], cum.max() + min_max_margin])

        plot_height += (n - 1) * 70

        fig.update_xaxes({'range': temp_min_max})
        fig.update_layout(title_text=title, title_x=0.15, font={'color': "#371ea3"}, template="none",
                          height=plot_height, margin={'t': 78, 'b': 71, 'r': 30})

        if show:
            fig.show(config=_theme.get_default_config())
        else:
            return fig



import unittest
from dalex._global_checks import global_check_import


class Global(unittest.TestCase):

    def setUp(self):
        pass

    def test(self):
        with self.assertRaises(ImportError):
            global_check_import("error_package", "error")
        with self.assertRaises(ImportWarning):
            global_check_import("dalex", "warning")
        global_check_import("shap", "test")
        global_check_import("statsmodels")
        global_check_import("scikit-learn")
        global_check_import("lime")

        with self.assertRaises(BaseException):
            global_check_import("sklearn", "this won't work")



import numpy as np
import pandas as pd
from copy import deepcopy
from warnings import warn


def check_columns_in_new_observation(new_observation, explainer):
    if not set(new_observation.columns).issubset(explainer.data):
        raise ValueError(
            "Columns in new observation does not match these in training dataset."
        )


def check_new_observation(new_observation, explainer):
    new_observation_ = deepcopy(new_observation)
    if isinstance(new_observation_, pd.Series):
        new_observation_ = new_observation_.to_frame().T
        new_observation_.columns = explainer.data.columns
    elif isinstance(new_observation_, np.ndarray):
        if new_observation_.ndim == 1:
            # make 1D array 2D
            new_observation_ = new_observation_.reshape((1, -1))
        elif new_observation_.ndim > 2:
            raise ValueError("Wrong new_observation dimension")
        elif new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_ = pd.DataFrame(new_observation_)
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, list):
        new_observation_ = pd.DataFrame(new_observation_).T
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, pd.DataFrame):
        if new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_.columns = explainer.data.columns
    else:
        raise TypeError(
            "new_observation must be a numpy.ndarray or pandas.Series or pandas.DataFrame"
        )

    if pd.api.types.is_bool_dtype(new_observation_.index):
        raise ValueError("new_observation index is of boolean type")

    return new_observation_


def check_variable_groups(variable_groups, explainer):
    variable_groups_copy = deepcopy(variable_groups)
    if variable_groups_copy is not None:
        if not isinstance(variable_groups_copy, dict):
            raise TypeError("'variable_groups' should be of class dict")

        wrong_names = np.empty(len(variable_groups_copy))
        for i, key in enumerate(variable_groups_copy):
            if isinstance(variable_groups_copy[key], list):
                variable_groups_copy[key] = np.array(variable_groups_copy[key])
            elif not isinstance(variable_groups_copy[key], np.ndarray):
                raise TypeError(
                    "'variable_groups' is a dict of lists of variables"
                )

            if not isinstance(variable_groups_copy[key][0], str):
                raise TypeError(
                    "'variable_groups' is a dict of lists of variables"
                )

            wrong_names[i] = np.in1d(variable_groups_copy[key], explainer.data.columns).all()

        wrong_names = not wrong_names.all()

        if wrong_names:
            raise ValueError(
                "You have passed wrong variables names in variable_groups argument. "
                "'variable_groups' is a dict of lists of numeric variables."
            )

    return variable_groups_copy


def check_random_state(random_state):
    if random_state is not None:
        np.random.seed(random_state)

    return random_state


def check_method_type(type, types, aliases=None):
    if isinstance(type, tuple):
        ret = type[0]
    elif isinstance(type, str):
        ret = type
    else:
        raise TypeError("'type' is not a string")

    if ret not in types:
        if aliases:
            if ret not in aliases:
                raise ValueError("'type' must be one of: {}".format(', '.join(types+tuple(aliases))))
            else:
                return aliases[ret]
        else:
            if ret not in types:
                raise ValueError("'type' must be one of: {}".format(', '.join(types)))
    else:
        return ret


def check_processes(processes):
    from multiprocessing import cpu_count
    if processes > cpu_count():
        warn("You have asked for too many processes. Truncated to the number of physical CPUs.")
        return cpu_count()
    else:
        return processes

def check_method_depend(depend_method, corr_method, agg_method):
    depend_method_types = ('assoc', 'pps')
    depend_method_aliases = {'association': 'assoc', "PPS": 'pps', 'stats': 'assoc'}
    corr_method_types = ('spearman', 'pearson', 'kendall')
    agg_method_types = ('max', 'min', 'mean')
    agg_method_aliases = {'maximum': 'max', 'minimum': 'min', 'avg': 'mean', 'average': 'mean'}
    if isinstance(depend_method, str):
        if depend_method not in depend_method_types:
            if depend_method not in depend_method_aliases:
                raise ValueError("'depend_method' must be one of: {}".format(', '.join(depend_method_types+tuple(depend_method_aliases))))
            else:
                depend_method = depend_method_aliases[depend_method]
        if depend_method == "assoc":
            if corr_method not in corr_method_types:
                raise ValueError("'corr_method' must be one of: {}".format(', '.join(corr_method_types)))
        if depend_method == "pps":
            if agg_method not in agg_method_types:
                if agg_method not in agg_method_aliases:
                    raise ValueError("'agg_method' must be one of: {}".format(', '.join(agg_method_types)))
                else: 
                    agg_method = agg_method_aliases[agg_method]
    return depend_method, corr_method, agg_method





import unittest

from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import LabelEncoder
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression, LinearRegression

import dalex as dx
import pandas as pd


class ModelSurrogateTest(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_titanic()
        self.X = data.drop(columns=['survived', 'class', 'embarked'])
        self.y = data.survived
        self.X.gender = LabelEncoder().fit_transform(self.X.gender)

        # this checks for no feature_importances_ attribute
        model = MLPClassifier(hidden_layer_sizes=(50, 50), max_iter=400, random_state=0)
        model.fit(self.X, self.y)
        self.exp = dx.Explainer(model, self.X, self.y, verbose=False)

        data2 = dx.datasets.load_fifa()
        self.X2 = data2.drop(["nationality", "overall", "potential",
                              "value_eur", "wage_eur"], axis=1).iloc[0:2000, 0:10]
        self.y2 = data2['value_eur'].iloc[0:2000]

        # this checks for feature_importances_ attribute
        model2 = RandomForestRegressor(random_state=0)
        model2.fit(self.X2, self.y2)
        self.exp2 = dx.Explainer(model2, self.X2, self.y2, verbose=False)

    def test(self):
        case1 = self.exp.model_surrogate()
        case2 = self.exp2.model_surrogate()
        case3 = self.exp.model_surrogate(max_vars=3, max_depth=2)
        case4 = self.exp2.model_surrogate(max_vars=3, max_depth=2)

        case5 = self.exp.model_surrogate(type='linear')
        case6 = self.exp2.model_surrogate(type='linear')
        case7 = self.exp.model_surrogate(type='linear', max_vars=3)
        case8 = self.exp2.model_surrogate(type='linear', max_vars=3)

        self.assertIsInstance(case1, DecisionTreeClassifier)
        self.assertIsInstance(case1.performance, pd.DataFrame)
        self.assertTrue(hasattr(case1, 'plot'))
        self.assertIsInstance(case2, DecisionTreeRegressor)
        self.assertIsInstance(case2.performance, pd.DataFrame)
        self.assertTrue(hasattr(case2, 'plot'))
        self.assertIsInstance(case3, DecisionTreeClassifier)
        self.assertIsInstance(case3.performance, pd.DataFrame)
        self.assertTrue(hasattr(case3, 'plot'))
        self.assertIsInstance(case4, DecisionTreeRegressor)
        self.assertIsInstance(case4.performance, pd.DataFrame)
        self.assertTrue(hasattr(case4, 'plot'))

        self.assertIsInstance(case5, LogisticRegression)
        self.assertIsInstance(case5.performance, pd.DataFrame)
        self.assertIsInstance(case6, LinearRegression)
        self.assertIsInstance(case6.performance, pd.DataFrame)
        self.assertIsInstance(case7, LogisticRegression)
        self.assertIsInstance(case7.performance, pd.DataFrame)
        self.assertIsInstance(case8, LinearRegression)
        self.assertIsInstance(case8.performance, pd.DataFrame)


if __name__ == '__main__':
    unittest.main()



import time
from . import plots
from tqdm import tqdm
from ._plot_container import PlotContainer


class PlotsManager:
    """Creates PlotsManager object

    This class should be only created by arena instance to manage its plots.

    Parameters
    ----------
    arena : dalex.Arena
        Instance of Arena.

    Attributes
    --------
    arena : Arena
        Instance of dalex.Arena
    cache : list of PlotContainer objects
        List of already calculated plots
    mutex : _thread.lock
        Mutex for params, plots and resources cache. Should be common with mutex from Arena instance.
    plots : list of classes extending PlotContainer
        List of available plot classes
    """
    def __init__(self, arena):
        if type(arena).__name__ != 'Arena' or type(arena).__module__ != 'dalex.arena.object':
            raise Exception('Invalid Arena argument')
        self.arena = arena
        self.cache = []
        self.mutex = arena.mutex
        self.plots = [vars(plots)[res] for res in getattr(plots, '__all__')]

    def get_supported_plots(self):
        """Returns plots classes that can produce at least one valid chart for parent arena.

        Returns
        -----------
        List of classes extending PlotContainer
        """
        return [plot for plot in self.plots if plot.test_arena(self.arena)]

    def clear_cache(self, plot_type=None):
        """Clears cache

        Parameters
        -----------
        plot_type : str or None
            If None all cache is cleared. Otherwise only plots with
            provided plot_type are removed.

        Notes
        -------
        This function must be called from mutex context
        """
        if plot_type is None:
            self.cache = []
        else:
            self.cache = list(filter(lambda p: p.plot_type != plot_type, self.cache))
        self.arena.update_timestamp()

    def find_in_cache(self, plot_type, params):
        """Function searches for cached plot

        Parameters
        -----------
        plot_type : str
            Value of plot_type field, that requested plot must have
        params : dict
            Keys of this dict are params types (model, observation, variable, dataset)
            and values are corresponding params labels. Requested plot must have equal
            params field.

        Returns
        --------
        PlotContainer or None
        """

        def _filter(p):
            return p.plot_type == plot_type and params == p.params
        with self.mutex:
            return next(filter(_filter, self.cache), None)

    def put_to_cache(self, plot_container):
        """Puts new plot to cache

        Parameters
        -----------
        plot_container : PlotContainer
        """
        if not isinstance(plot_container, PlotContainer):
            raise Exception('Invalid plot container')
        with self.mutex:
            self.cache.append(plot_container)

    def fill_cache(self, fixed_params={}):
        """Generates all available plots and cache them

        This function tries to generate all plots that are not cached already and
        put them to cache. Range of generated plots can be narrow using `fixed_params`

        Parameters
        -----------
        fixed_params : dict
            This dict specifies which plots should be generated. Only those with
            all keys from `fixed_params` present and having the same value will be
            calculated.
        """
        if not isinstance(fixed_params, dict):
            raise Exception('Params argument must be a dict')
        for plot_class in self.get_supported_plots():
            required_params = plot_class.info.get('requiredParams')
            # Test if all params fixed by user are used in this plot. If not, then skip it.
            # This list contains fixed params' types, that are not required by plot.
            # Loop will be skipped if this list is not empty.
            if len([k for k in fixed_params.keys() if k not in required_params]) > 0:
                continue
            available_params = self.arena.get_available_params()
            iteration_pools = map(lambda p: available_params.get(p) if fixed_params.get(p) is None else [fixed_params.get(p)], required_params)
            combinations = [[]]
            for pool in iteration_pools:
                combinations = [x + [y] for x in combinations for y in pool]
            if self.arena.verbose and len(combinations) > 0:
                combinations = tqdm(combinations)
                combinations.set_description(plot_class.info.get('name'))
            for params_values in combinations:
                params = dict(zip(required_params, params_values))
                self.get_plot(plot_type=plot_class.info.get('plotType'), params_values=params, wait=True)

    def get_plot(self, plot_type, params_values, cache=True, wait=False):
        """Returns plot for specified type and params

        Function serches for plot in cache, when not present creates
        requested plot and put it to cache.

        Parameters
        -----------
        plot_type : str
            Type of plot to be generated
        params_values : dict
            Dict for param types as keys and Param objects as values
        cache : bool
            If serach for plot in cache and put calculated plot into cache.

        Returns
        --------
        PlotContainer
        """
        plot_class = next((c for c in self.plots if c.info.get('plotType') == plot_type), None)
        if plot_class is None:
            raise Exception('Not supported plot type')
        plot_type = plot_class.info.get('plotType')
        required_params_values = {}
        required_params_labels = {}
        for p in plot_class.info.get('requiredParams'):
            if params_values.get(p) is None:
                raise Exception('Required param is missing')
            required_params_values[p] = params_values.get(p)
            required_params_labels[p] = params_values.get(p).get_label()
        result = self.find_in_cache(plot_type, required_params_labels) if cache else None
        if result is None:
            result = plot_class(self.arena, cache=cache).fit(required_params_values)
            while wait and result.is_done == False:
                time.sleep(0.5)
                result = plot_class(self.arena, cache=cache).fit(required_params_values)
            if cache and result.is_done != False:
                self.put_to_cache(result)
        return result



import numpy as np
import plotly.graph_objects as go
from copy import deepcopy 

from dalex import _global_utils, _theme

from . import utils



def plot_predict_hierarchical_importance(
    dendrogram_hierarchical_correlation,
    hierarchical_importance_data,
    rounding_function,
    digits,
    absolute_value,
    type
):
    dendrogram_hierarchical_importance = go.Figure(dendrogram_hierarchical_correlation)
    importance_dict = {}
    min_imp = min(hierarchical_importance_data["importance"])
    max_imp = max(hierarchical_importance_data["importance"])


    tick_dict = dict(zip(dendrogram_hierarchical_importance.layout.yaxis.tickvals, 
                    [[var] for var in dendrogram_hierarchical_importance.layout.yaxis.ticktext]))
    aspects_dendro_order = []
    aspects_vars_before = []
    for scatter in dendrogram_hierarchical_importance.data:
        tick_dict[np.mean(scatter.y[1:3])] = tick_dict[scatter.y[1]] + tick_dict[scatter.y[2]]
        aspects_dendro_order.append(tick_dict[scatter.y[1]] + tick_dict[scatter.y[2]])
        aspects_vars_before.append((tick_dict[scatter.y[1]], tick_dict[scatter.y[2]]))

    updated = []
    for i, row in hierarchical_importance_data[:-1].iterrows():
        dendro_mask = np.array([set(row.variable_names) == set(el) for el in aspects_dendro_order])
        if dendro_mask.any():
            k = np.flatnonzero(dendro_mask == True)[0]
            updated.append(k)
            imp_scatter = row["importance"]
            if absolute_value:
                imp_scatter = abs(imp_scatter)
            scatter_importance = dendrogram_hierarchical_importance.data[k]
            scatter_clustering = dendrogram_hierarchical_correlation.data[k]
            importance_dict[tuple(aspects_dendro_order[k])] = imp_scatter
            scatter_importance.x[1:3] = imp_scatter

            scatter_importance.hoverinfo = "text"
            scatter_importance.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter_importance.hovertext = tooltip_text_aspect(row, rounding_function, digits)

            scatter_clustering.hoverinfo = "text"
            scatter_clustering.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter_clustering.hovertext = tooltip_text_aspect(row, rounding_function, digits)

    for i, scatter in enumerate(dendrogram_hierarchical_importance.data):
        vars_before = aspects_vars_before[i]
        if i not in updated:
            if absolute_value == True:
                imp_scatter = abs(hierarchical_importance_data.iloc[-1]["importance"])
            else:
                imp_scatter = hierarchical_importance_data.iloc[-1]["importance"]
            change_last = True if type == 'default' else False
            scatter.hoverinfo = "text"
            scatter.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter.hovertext = tooltip_text_aspect(hierarchical_importance_data.iloc[-1], rounding_function, digits, change_last)
            scatter_clustering = dendrogram_hierarchical_correlation.data[i]
            scatter_clustering.hoverinfo = "text"
            scatter_clustering.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter_clustering.hovertext = tooltip_text_aspect(hierarchical_importance_data.iloc[-1], rounding_function, digits, change_last)
            scatter.x[1:3] = imp_scatter
            importance_dict[tuple(aspects_dendro_order[i])] = imp_scatter
        if tuple(vars_before[0]) in importance_dict.keys():
            scatter.x[0] = importance_dict[tuple(vars_before[0])]
        if tuple(vars_before[1]) in importance_dict.keys():
            scatter.x[3] = importance_dict[tuple(vars_before[1])]

    min_range = (
        min_imp * 1.15
        if min_imp < 0
        else min_imp * 0.85
    )
    max_range = (
        max_imp * 1.15
        if max_imp * 1.15 > 0
        else max_imp * 0.85
    )
    dendrogram_hierarchical_importance.update_xaxes(
        range=[min_range, max_range],
        visible=False,
        showticklabels=False,
    )
    return dendrogram_hierarchical_importance, updated


def get_ticktext_for_plot(
    single_aspect_importance_df, variables_order, abbrev_labels=0
):
    _df = deepcopy(single_aspect_importance_df)
    ticktext = [
        f"{variable} = {utils.nice_format(_df.loc[_df['variable_names']==variable, 'variable_values'].values[0])}"
        if abbrev_labels <= 0
        else utils.text_abbreviate(
            f"{variable} = {utils.nice_format(_df.loc[_df['variable_names']==variable, 'variable_values'].values[0])}",
            abbrev_labels,
        )
        for variable in variables_order
    ]
    return ticktext


def tooltip_text_aspect(row, rounding_function, digits, last=False):
    var_val_string = ""
    for i in range(len(row.variable_names)):
        var_val_string += (
            "<br>" + row.variable_names[i] + " = " + str(row.variable_values[i])
        )
    keyword = "Change in average response: " if last else "Importance: " 

    return (
        f"Min abs depend: {rounding_function(row.min_depend, digits)}<br>"
        + "(between variables: " + ", ".join(row.vars_min_depend) + ")<br>"
        + keyword + f"{rounding_function(row.importance, digits)}<br>Variables: "
        + var_val_string
    )


def tooltip_text_single(row, rounding_function, digits):
    return (
        f"Importance: {rounding_function(row.importance, digits)}<br>"
        + "Variable:<br>"
        + f"{row.variable_names} = " + str(row.variable_values)
    )


def add_text_to_dendrogram(fig, updated, rounding_function, digits, type="clustering"):
    res_fig = go.Figure(fig)
    corner_scatters = [] 
    for i, scatter in enumerate(res_fig.data):
        if i in updated:
            x_cord = scatter.x[1]
            y_cord = np.mean(scatter.y[1:3])
            if type == "clustering":
                lab_text = str(rounding_function(1-x_cord, digits))
            else:
                lab_text = str(rounding_function(x_cord, digits))
            scatter.x = np.insert(scatter.x, 2, x_cord)
            scatter.y = np.insert(scatter.y, 2, y_cord)
            scatter.mode = "text+lines"
            scatter["text"] = [None, None, lab_text, None, None]
            scatter["textposition"] = "middle left"
        else:
            corner_scatters.append(scatter)
    if corner_scatters:
        scatter = corner_scatters[len(corner_scatters)//2]
        x_cord = scatter.x[1]
        y_cord = np.mean(scatter.y[1:3])
        if type == "clustering":
            lab_text = str(rounding_function(1-x_cord, digits))
        else:
            lab_text = str(rounding_function(x_cord, digits))
        scatter.x = np.insert(scatter.x, 2, x_cord)
        scatter.y = np.insert(scatter.y, 2, y_cord)
        scatter.mode = "text+lines"
        scatter["text"] = [None, None, lab_text, None, None]
        scatter["textposition"] = "middle left"

    return res_fig


def plot_single_aspects_importance(
    result_dataframe, order, rounding_function, digits, vcolors
):
    fig = go.Figure()
    _result = deepcopy(result_dataframe)
    sorter = dict(zip(order, range(len(order))))
    _result["order"] = _result["variable_names"].map(sorter)
    _result = _result.sort_values(["order"], ascending=True).reset_index(drop=True)
    _result = _result.drop("order", axis=1)
    _result.loc[:, "importance"] = rounding_function(
                _result.loc[:, "importance"], digits
            )

    _result["color"] = [0 if imp > 0 else 1 for imp in _result["importance"]]
    _result["tooltip_text"] = _result.apply(tooltip_text_single, args=(rounding_function, digits), axis=1)
    _result["label_text"] = _global_utils.convert_float_to_str(_result.importance, "+")

    if vcolors is None:
        vcolors = _theme.get_aspect_importance_colors()

    fig.add_shape(
        type="line",
        x0=0,
        x1=0,
        y0=-0.01,
        y1=1.01,
        yref="paper",
        xref="x",
        line={"color": "#371ea3", "width": 1.5, "dash": "dot"},
    )

    fig.add_bar(
        orientation="h",
        y=_result["variable_names"].tolist(),
        x=_result["importance"].tolist(),
        textposition="outside",
        text=_result["label_text"].tolist(),
        textfont_color=["#371ea3"] * len(order),
        marker_color=[vcolors[int(c)] for c in _result["color"].tolist()],
        hovertext=_result["tooltip_text"].tolist(),
        hoverinfo="text",
        hoverlabel={"bgcolor": "rgba(0,0,0,0.8)"},
        showlegend=False,
    )

    fig.update_yaxes(
        {
            "type": "category",
            "autorange": "reversed",
            "gridwidth": 2,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 10,
            "fixedrange": True,
        }
    )

    fig.update_xaxes(
        {
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
        }
    )

    temp_min_max = [np.Inf, -np.Inf]
    min_max_margin = _result.importance.values.ptp() * 0.15
    temp_min_max[0] = np.min(
        [temp_min_max[0], _result.importance.values.min() - min_max_margin]
    )
    temp_min_max[1] = np.max(
        [temp_min_max[1], _result.importance.values.max() + min_max_margin]
    )

    fig.update_xaxes({"range": temp_min_max})
    fig.update_layout(
        title_x=0.15,
        font={"color": "#371ea3"},
        template="none",
        margin={"t": 78, "b": 71, "r": 30},
    )

    return fig


def _add_between_points(
        arr, arr_2, val_arr_2, range_global, i_begin, i_end, arr_text=None
    ):
        curr_val = arr[i_begin]
        last_val = arr[i_end]
        range_curr = abs(curr_val - last_val)
        num = int((range_curr * 40) / range_global)
        num = max(num, 5)
        new_points = np.linspace(
            curr_val,
            last_val,
            num=num,
        )
        i_curr = i_begin + 1
        arr = np.insert(arr, i_curr, new_points[1:-1])
        arr_2 = np.insert(arr_2, i_curr, [val_arr_2] * (num - 2))
        if arr_text is not None:
            arr_text = np.insert(arr_text, i_curr, [None] * (num - 2))
        return arr, arr_2, num - 2, arr_text

def _add_points_on_dendrogram_traces(fig):
    k = len(fig.data)
    range_x2 = fig.full_figure_for_development(warn=False).layout.xaxis2.range
    range_x2_len = range_x2[1] - range_x2[0]
    range_x3 = fig.full_figure_for_development(warn=False).layout.xaxis3.range
    range_x3_len = range_x3[1] - range_x3[0]
    range_y = fig.full_figure_for_development(warn=False).layout.yaxis.range
    range_y_len = range_y[1] - range_y[0]
    middle_point = [None] * k
    for i in range(1, k):
        if i < ((k - 1) / 2 + 1):
            range_x_len = range_x2_len
        else:
            range_x_len = range_x3_len
        arr_text = None
        num_points = len(fig.data[i]["x"])
        if num_points == 5:
            arr_text = fig.data[i]["text"]
        middle_point[i] = fig.data[i]["x"][2], fig.data[i]["y"][2]
        inserted_points = 0
        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        fig.data[i]["y"], fig.data[i]["x"], j, arr_text = _add_between_points(
            fig.data[i]["y"],
            fig.data[i]["x"],
            fig.data[i]["x"][inserted_points],
            range_y_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        if num_points == 5:
            (
                fig.data[i]["y"],
                fig.data[i]["x"],
                j,
                arr_text,
            ) = _add_between_points(
                fig.data[i]["y"],
                fig.data[i]["x"],
                fig.data[i]["x"][inserted_points],
                range_y_len,
                inserted_points,
                inserted_points + 1,
                arr_text,
            )
            inserted_points += j + 1
        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        
        if arr_text is not None:
            fig.data[i]["text"] = arr_text
    return fig, middle_point




import pandas as pd
import numpy as np
from .._plot_container import PlotContainer

class ROCContainer(PlotContainer):
    info = {
        'name': 'Receiver Operating Characterstic',
        'plotType': 'ROC',
        'plotCategory': 'Model Performance',
        'requiredParams': ['model']
    }
    options_category = 'ROC'
    options = {
        'grid_points': { 'default': 101, 'desc': 'Maximum number of points for ROC curve' },
    }

    def _fit(self, model):
        exp = model.explainer
        if exp.model_type != 'classification':
            self.set_message('ROC plot is only available for classificators')
            return

        y_hat = exp.predict(exp.data) if exp.y_hat is None else exp.y_hat
        df = pd.DataFrame({ 'y': exp.y.astype(bool), 'y_hat': y_hat })

        P_n = df.y.sum()
        N_n = df.shape[0] - P_n
        if P_n == 0 or N_n == 0:
            self.set_message('Provided dataset contains only positive or only negative cases.', 'error')
            return

        tpr_temp = df.groupby('y_hat').sum().reset_index().sort_values('y_hat', ascending=False)
        fpr_temp = df.assign(y=1-df.y).groupby('y_hat').sum().reset_index().sort_values('y_hat', ascending=False)
        
        _df = pd.DataFrame({
            'TPR': tpr_temp.y.cumsum() / P_n,
            'TNR': 1 - (fpr_temp.y.cumsum() / N_n),
            'cutoff': np.sort(df['y_hat'].unique())[::-1]
        })

        grid_points = self.get_option('grid_points')
        if _df.shape[0] > grid_points:
            _df = _df.sample(grid_points).sort_values('cutoff', ascending=False)

        self.data = {
            'cutoff': _df['cutoff'].tolist(),
            'specifity': [1] + _df['TNR'].tolist(),
            'sensivity': [0] + _df['TPR'].tolist()
        }

    def test_arena(arena):
        if type(arena).__name__ != 'Arena' or type(arena).__module__ != 'dalex.arena.object':
            raise Exception('Invalid Arena argument')
        return next((True for model in arena.get_params('model') if model.explainer.model_type == 'classification'), False)



from .._plot_container import PlotContainer

class BreakDownContainer(PlotContainer):
    info = {
        'name': "Break Down",
        'plotType': "Breakdown",
        'plotCategory': "Observation Level",
        'requiredParams': ["model", "observation"]
    }
    options_category = 'BreakDown'
    options = {
    }
    def _fit(self, model, observation):
        row = observation.get_row_for_model(model)
        if row is None:
            self.set_message('Observation is not valid for given model.')
            return
        bd = model.explainer.predict_parts(row, type='break_down').result
        self.data = {
            'variables': bd[1:-1].variable_name.tolist(),
            'variables_value': bd[1:-1].variable_value.tolist(),
            'contribution': bd[1:-1].contribution.tolist(),
            'intercept': bd.contribution[0],
            'prediction': bd.cumulative.tail(1).iloc[0]
        }



import unittest
from copy import deepcopy

import numpy as np
import pandas as pd
from copy import copy
from plotly.graph_objs import Figure
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import DecisionTreeRegressor

import dalex as dx
from dalex.fairness import reweight, resample, roc_pivot


class FairnessTest(unittest.TestCase):

    def setUp(self):
        self.protected = np.array(['a', 'a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c', 'c'])
        self.y_true = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0])
        self.y_pred = np.array([0.32, 0.43, 0.56, 0.67, 0.9, 0.67, 0.98, 0.1, 0.44, 1, 0.65, 0.55, 1])
        self.cutoff = {'a': 0.5, 'b': 0.4, 'c': 0.6}

        # classifier
        data = dx.datasets.load_german()

        X = data.drop(columns='risk')
        y = data.risk

        categorical_features = ['sex', 'job', 'housing', 'saving_accounts', "checking_account", 'purpose']
        categorical_transformer = Pipeline(steps=[
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('cat', categorical_transformer, categorical_features)])

        clf = Pipeline(steps=[('preprocessor', preprocessor),
                              ('classifier', DecisionTreeClassifier(max_depth=7, random_state=123))])
        clf.fit(X, y)

        self.exp = dx.Explainer(clf, X, y, verbose=False)
        self.german_protected = data.sex + '_' + np.where(data.age < 25, 'young', 'old')

        self.mgf = self.exp.model_fairness(protected=self.german_protected,
                                           privileged='male_old',
                                           verbose=False)
        self.mgf2 = deepcopy(self.mgf)
        self.mgf.label = 'first'
        self.mgf2.label = 'second'

        # regressor
        self.protected_reg = np.array([np.tile('A', 1000), np.tile('B', 1000)]).flatten()
        first = np.array([np.random.normal(100, 20, 1000), np.random.normal(50, 10, 1000)]).flatten()
        second = np.array([np.random.normal(60, 20, 1000), np.random.normal(60, 10, 1000)]).flatten()
        target = np.array([np.random.normal(10000, 2000, 1000), np.random.normal(8000, 1000, 1000)]).flatten()
        data2 = pd.DataFrame({'first': first, 'second': second})

        reg = DecisionTreeRegressor()
        reg.fit(data2, target)

        self.exp_reg = dx.Explainer(reg, data2, target)
        self.mgf_reg = self.exp_reg.model_fairness(self.protected_reg, 'A')

    def test_fairness_check(self):
        self.mgf.fairness_check()
        self.mgf2.fairness_check()
        self.mgf_reg.fairness_check()

        self.mgf.fairness_check(epsilon=0.1)
        self.mgf_reg.fairness_check(epsilon=0.1)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            self.mgf.fairness_check(epsilon=-1)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            self.mgf_reg.fairness_check(epsilon=-1)

    def test_ConfusionMatrix(self):
        y_true = np.array([0, 0, 0, 0, 1, 1, 1, 1])
        y_pred = np.array([0.32, 0.54, 0.56, 0.67, 0.34, 0.67, 0.98, 1])
        cutoff = 0.55
        cm = dx.fairness._group_fairness.utils.ConfusionMatrix(y_true, y_pred, cutoff)

        #  proper calculations
        self.assertEqual(cm.cutoff, 0.55)
        self.assertEqual(cm.tp, 3)
        self.assertEqual(cm.tn, 2)
        self.assertEqual(cm.fp, 2)
        self.assertEqual(cm.fn, 1)

        #  error assertions
        y_true = y_true[:-1]
        with self.assertRaises(AssertionError):
            cm_ = dx.fairness._group_fairness.utils.ConfusionMatrix(y_true, y_pred, cutoff)
        y_true = np.append(y_true, 1)
        cutoff = 1.5
        with self.assertRaises(AssertionError):
            cm_ = dx.fairness._group_fairness.utils.ConfusionMatrix(y_true, y_pred, cutoff)

    def test_SubConfusionMatrix(self):
        y_true = self.y_true.copy()
        y_pred = self.y_pred
        protected = self.protected
        cutoff = self.cutoff

        scf = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)

        #  proper calculations
        self.assertEqual(scf.sub_dict.get('a').tp, 2)
        self.assertEqual(scf.sub_dict.get('a').fp, 2)
        self.assertEqual(scf.sub_dict.get('a').fn, 0)
        self.assertEqual(scf.sub_dict.get('a').tn, 2)

        self.assertEqual(scf.sub_dict.get('b').tp, 2)
        self.assertEqual(scf.sub_dict.get('b').fp, 0)
        self.assertEqual(scf.sub_dict.get('b').fn, 1)
        self.assertEqual(scf.sub_dict.get('b').tn, 0)

        self.assertEqual(scf.sub_dict.get('c').tp, 2)
        self.assertEqual(scf.sub_dict.get('c').fp, 1)
        self.assertEqual(scf.sub_dict.get('c').fn, 0)
        self.assertEqual(scf.sub_dict.get('c').tn, 1)

        #  error assertions
        y_true = y_true[:-1]
        with self.assertRaises(AssertionError):
            cm_ = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)
        y_true = np.append(y_true, 1)
        cutoff = [0.1, 0.2, 0.4]  # list instead of dict
        with self.assertRaises(AssertionError):
            cm_ = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)

    def test_SubgroupConfusionMatrixMetrics(self):
        y_true = self.y_true
        y_pred = self.y_pred
        protected = self.protected
        cutoff = self.cutoff

        scf = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)
        scf_metrics = dx.fairness._group_fairness.utils.SubgroupConfusionMatrixMetrics(scf)
        scmm = scf_metrics.subgroup_confusion_matrix_metrics
        self.assertEqual(scmm.get('a').get('TPR'), 1)
        self.assertEqual(scmm.get('b').get('TPR'), 0.667)
        self.assertTrue(np.isnan(scmm.get('b').get('TNR')))

    def test_calculate_ratio(self):
        y_true = self.y_true
        y_pred = self.y_pred
        protected = self.protected
        cutoff = self.cutoff

        scf = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)
        scf_metrics = dx.fairness._group_fairness.utils.SubgroupConfusionMatrixMetrics(scf)

        df_ratio = dx.fairness._group_fairness.utils.calculate_ratio(scf_metrics, 'a')

        b = list(scf_metrics.subgroup_confusion_matrix_metrics.get('b').values())
        a = list(scf_metrics.subgroup_confusion_matrix_metrics.get('a').values())

        ratio = np.array(b) / np.array(a)
        ratio[np.isinf(ratio)] = np.nan
        ratio[ratio == 0] = np.nan

        ratio_nonnan = ratio[np.isfinite(ratio)]
        df_ratio_nonnan = np.array(df_ratio.iloc[1, :][np.isfinite(df_ratio.iloc[1, :])])

        self.assertTrue(np.equal(ratio_nonnan, df_ratio_nonnan).all())

    def test_calculate_parity_loss(self):
        y_true = self.y_true
        y_pred = self.y_pred
        protected = self.protected
        cutoff = self.cutoff

        scf = dx.fairness._group_fairness.utils.SubgroupConfusionMatrix(y_true, y_pred, protected, cutoff)
        scf_metrics = dx.fairness._group_fairness.utils.SubgroupConfusionMatrixMetrics(scf)

        parity_loss = dx.fairness._group_fairness.utils.calculate_parity_loss(scf_metrics, "a")
        ratios = dx.fairness._group_fairness.utils.calculate_ratio(scf_metrics, "a")
        TPR_parity_loss = parity_loss.iloc[0]

        TPR_ratios = ratios.TPR / ratios.TPR[0]
        TPR_log = abs(np.log(TPR_ratios))

        self.assertEqual(TPR_log.sum(), TPR_parity_loss)

    def test_GroupFairnessClassification(self):
        exp = self.exp
        protected = self.german_protected

        gfco = dx.fairness.GroupFairnessClassification(y=exp.y,
                                                       y_hat=exp.y_hat,
                                                       protected=protected,
                                                       privileged='male_old',
                                                       verbose=False,
                                                       label=exp.label)
        self.assertIsInstance(gfco, dx.fairness.GroupFairnessClassification)

    def test_GroupFairnessRegression(self):
        exp = self.exp_reg
        protected = self.protected_reg

        gfco = dx.fairness.GroupFairnessRegression(y=exp.y,
                                                   y_hat=exp.y_hat,
                                                   protected=protected,
                                                   privileged='A',
                                                   verbose=False,
                                                   label=exp.label)
        self.assertIsInstance(gfco, dx.fairness.GroupFairnessRegression)

    def test_parameter_checks(self):
        exp = self.exp
        protected = self.german_protected

        #  error handling
        wrong_protected = np.array([protected, protected])
        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = exp.model_fairness(protected=wrong_protected,
                                      privileged='male_old',
                                      verbose=False)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = exp.model_fairness(protected=protected,
                                      privileged='not_existing',
                                      verbose=False)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = exp.model_fairness(protected=protected,
                                      privileged='male_old',
                                      epsilon=1.1,  # wrong epsilon
                                      verbose=False)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = dx.fairness.GroupFairnessClassification(y=exp.y[:-1, ],
                                                           y_hat=exp.y_hat,
                                                           protected=protected,
                                                           privileged='male_old',
                                                           verbose=False,
                                                           label=exp.label)
        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = dx.fairness.GroupFairnessClassification(y=exp.y[:-1, ],
                                                           y_hat=exp.y_hat[:-1, ],
                                                           protected=protected,
                                                           privileged='male_old',
                                                           verbose=False,
                                                           label=exp.label)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = exp.model_fairness(protected=protected,
                                      privileged='male_old',
                                      cutoff=1.2,
                                      verbose=False)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            gfco = exp.model_fairness(protected=protected,
                                      privileged='male_old',
                                      cutoff='not_int',
                                      verbose=False)

        # conversion check
        gfco = exp.model_fairness(protected=protected,
                                  privileged='male_old',
                                  cutoff=0.6,
                                  verbose=False)

        self.assertEqual(list(gfco.cutoff.values()), [0.6, 0.6, 0.6, 0.6])

        gfco = exp.model_fairness(protected=protected,
                                  privileged='male_old',
                                  cutoff={'male_old': 0.9},
                                  verbose=False)
        self.assertEqual(gfco.cutoff, {'male_old': 0.9, 'female_old': 0.5, 'male_young': 0.5, 'female_young': 0.5})

        np.random.seed(1)
        new_protected = np.random.choice(np.array([0, 1]), 1000)
        gfco = exp.model_fairness(protected=new_protected,
                                  privileged=1,
                                  verbose=False)

        self.assertEqual(gfco.privileged, '1')
        self.assertEqual(list(gfco.protected), list(new_protected.astype('U')))

        gfco = exp.model_fairness(protected=list(protected),
                                  privileged='male_old',
                                  verbose=False)

        self.assertEqual(type(gfco.protected), np.ndarray)
        self.assertIsInstance(gfco, dx.fairness.GroupFairnessClassification)

    def test_model_group_fairness(self):
        exp = self.exp
        protected = self.german_protected
        mgf = exp.model_fairness(protected=protected,
                                 privileged='male_old',
                                 verbose=False)

        exp2 = self.exp_reg
        protected2 = self.protected_reg
        mgf2 = exp2.model_fairness(protected=protected2,
                                 privileged='A',
                                 verbose=False)

        self.assertIsInstance(mgf, dx.fairness.GroupFairnessClassification)
        self.assertIsInstance(mgf2, dx.fairness.GroupFairnessRegression)

    def test_plot_error_handling(self):
        # handling same for all plots, here fairness check plot
        with self.assertRaises(dx.fairness._basics.exceptions.FairnessObjectsDifferenceError):
            mgf_wrong = self.exp.model_fairness(protected=self.german_protected,
                                                privileged='male_young',  # wrong privileged
                                                verbose=False
                                                )
            self.mgf.plot([mgf_wrong])

        with self.assertRaises(TypeError):
            self.mgf.plot(objects= [self.mgf_reg])

        with self.assertRaises(TypeError):
            self.mgf_reg.plot(objects=[self.mgf])

        with self.assertRaises(dx.fairness._basics.exceptions.FairnessObjectsDifferenceError):
            mgf_wrong = self.exp.model_fairness(protected=self.german_protected,
                                                privileged='male_old',
                                                epsilon=0.6,  # different epsilon
                                                verbose=False)
            self.mgf.plot([mgf_wrong])

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            mgf_wrong = self.exp.model_fairness(protected=self.german_protected,
                                                privileged='male_old',
                                                epsilon=0.8,  # here ok epsilon
                                                verbose=False)
            self.mgf.plot([mgf_wrong], epsilon=1.2)  # here wrong epsilon

        with self.assertRaises(dx.fairness._basics.exceptions.FairnessObjectsDifferenceError):
            exp_wrong = deepcopy(self.exp)
            exp_wrong.y = exp_wrong.y[:-1]
            exp_wrong.y_hat = exp_wrong.y_hat[:-1]

            mgf_wrong = exp_wrong.model_fairness(protected=self.german_protected[:-1],  # shorter protected
                                                 privileged='male_old',
                                                 verbose=False)
            self.mgf.plot([mgf_wrong])

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            self.mgf.plot(type='not_existing')

    def test_plot_fairness_check(self):
        fig1 = self.mgf.plot(show=False)
        self.assertEqual(fig1.layout.title.text, "Fairness Check")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False)
        self.assertIsInstance(fig2, Figure)

        self.assertEqual(fig2['data'][0]['legendgroup'], "first")
        self.assertEqual(fig2['data'][5]['legendgroup'], "second")

        fig3 = self.mgf_reg.plot(show=False)
        self.assertEqual(fig3.layout.title.text, "Fairness Check")
        self.assertIsInstance(fig3, Figure)

        fig4 = self.mgf.plot(objects=self.mgf2, show=False)
        self.assertIsInstance(fig4, Figure)

        fig5 = self.mgf.plot(objects=self.mgf2, show=False, title='Test title')
        self.assertEqual(fig5.layout.title.text, 'Test title')



    def test_plot_metric_scores(self):
        fig1 = self.mgf.plot(show=False, type='metric_scores')
        self.assertEqual(fig1.layout.title.text, "Metric Scores")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='metric_scores')
        self.assertIsInstance(fig2, Figure)

        self.assertEqual(fig2['data'][0]['legendgroup'], "first")
        self.assertEqual(fig2['data'][5]['legendgroup'], "second")

    def test_plot_radar(self):
        fig1 = self.mgf.plot(show=False, type='radar')
        self.assertEqual(fig1.layout.title.text, "Fairness Radar")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='radar')
        self.assertIsInstance(fig2, Figure)

    def test_plot_heatmap(self):
        fig1 = self.mgf.plot(show=False, type='heatmap')
        self.assertEqual(fig1.layout.title.text, "Fairness Heatmap")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='heatmap')
        self.assertIsInstance(fig2, Figure)

    def test_plot_stacked(self):
        fig1 = self.mgf.plot(show=False, type='stacked')
        self.assertEqual(fig1.layout.title.text, "Stacked Parity Loss Metrics")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='stacked')
        self.assertIsInstance(fig2, Figure)

    def test_plot_performance_and_fairness(self):
        fig1 = self.mgf.plot(show=False, type='performance_and_fairness')
        self.assertEqual(fig1.layout.title.text, "Performance and Fairness")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='performance_and_fairness')
        self.assertIsInstance(fig2, Figure)

    def test_plot_ceteris_paribus_cutoff(self):
        fig1 = self.mgf.plot(show=False, type='ceteris_paribus_cutoff', subgroup='male_old')
        self.assertEqual(fig1.layout.title.text, "Ceteris Paribus Cutoff")
        self.assertIsInstance(fig1, Figure)

        fig2 = self.mgf.plot(objects=[self.mgf2], show=False, type='ceteris_paribus_cutoff', subgroup='male_old')
        self.assertIsInstance(fig2, Figure)

    def test_plot_density(self):
        fig = self.mgf_reg.plot(show=False, type='density')
        self.assertEqual(fig.layout.title.text, "Density plot")

        self.assertIsInstance(fig, Figure)

    def test_mitigation_reweight(self):
        predicted_weights = reweight(np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0]), np.array([1, 1, 1, 1, 0, 0, 0, 1, 0, 1]))
        # actual weights as in article
        actual_weights = np.array([0.75, 0.75, 0.75, 0.75, 2, 0.67, 0.67, 1.5, 0.67, 1.5])

        self.assertTrue(np.all(np.around(predicted_weights.astype(float), 2) == actual_weights))

    def test_mitigation_resample(self):
        # uniform
        df = pd.DataFrame({'sex':np.concatenate((np.repeat("M", 5), np.repeat("F", 5), np.repeat("N", 5)), axis = 0),
                           'target': [1, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1]})

        MN = sum((df.sex == "M") & (df.target == 0))
        MP = sum((df.sex == "M") & (df.target == 1))
        FN = sum((df.sex == "F") & (df.target == 0))
        FP = sum((df.sex == "F") & (df.target == 1))
        NN = sum((df.sex == "N") & (df.target == 0))
        NP = sum((df.sex == "N") & (df.target == 1))

        weights = reweight(df.sex, df.target)

        wMP = weights[0]
        wMN = weights[4]
        wFP = weights[9]
        wFN = weights[5]
        wNN = weights[12]
        wNP = weights[14]

        # expected
        E_MP = round(MP * wMP)
        E_MN = round(MN * wMN)
        E_FN = round(FN * wFN)
        E_FP = round(FP * wFP)
        E_NP = round(NP * wNP)
        E_NN = round(NN * wNN)

        # uniform
        df_2 = df.iloc[resample(df.sex, df.target), :]

        MN_2 = sum((df_2.sex == "M") & (df_2.target == 0))
        MP_2 = sum((df_2.sex == "M") & (df_2.target == 1))
        FN_2 = sum((df_2.sex == "F") & (df_2.target == 0))
        FP_2 = sum((df_2.sex == "F") & (df_2.target == 1))
        NN_2 = sum((df_2.sex == "N") & (df_2.target == 0))
        NP_2 = sum((df_2.sex == "N") & (df_2.target == 1))

        self.assertEqual(E_MP, MP_2)
        self.assertEqual(E_MN, MN_2)
        self.assertEqual(E_FP, FP_2)
        self.assertEqual(E_FN, FN_2)
        self.assertEqual(E_NP, MP_2)
        self.assertEqual(E_NN, NN_2)

        # preferential

        df = pd.DataFrame({'sex': np.concatenate((np.repeat("M", 5), np.repeat("F", 5), np.repeat("N", 5)), axis=0),
                           'target': [1, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1],
                           'name': np.arange(1,16),
                           'probs': [0.9, 0.82, 0.56, 0.78, 0.45, 0.12, 0.48, 0.63, 0.48, 0.88, 0.34, 0.12, 0.34, 0.49, 0.9]})


        df_3 = df.iloc[resample(df.sex, df.target, type = "preferential", probs = df.probs), :]
        self.assertTrue(np.all(sorted(np.array(df_3.name)) == np.array([1, 2, 5, 5, 5, 6, 7, 8, 9, 10, 11, 12, 13, 15, 15])))

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            resample(df.sex, df.target, type = "preferential", probs = [0,1,2,3,4])

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            resample(df.sex, df.target, type = "not_existing", probs = df.probs)

    def test_mitigation_roc(self):
        exp = copy(self.exp)
        exp.y_hat = np.array([0.3, 0.4, 0.45, 0.55, 0.6, 0.7])
        exp.y     = np.array([0, 0, 0, 1, 1, 1])
        protected = ['A', 'A', 'A', 'B', 'B', 'B']
        privileged = 'B'

        exp2 = roc_pivot(exp, protected, privileged, cutoff=0.5, theta=0.09)
        self.assertEqual(list(np.round(exp2.y_hat, 2)), [0.3, 0.4, 0.55, 0.45, 0.6, 0.7])

        exp.y_hat = np.array([0.3, 0.4, 0.45, 0.55, 0.6, 0.7])
        exp.y = np.array([0, 0, 0, 1, 1, 1])
        protected = ['A', 'A', 'A', 'B', 'B', 'B']
        privileged = 'B'

        exp2 = roc_pivot(exp, protected, privileged, cutoff=0.5, theta=0.11)
        self.assertEqual(list(np.round(exp2.y_hat, 2)), [0.3, 0.6, 0.55, 0.45, 0.4, 0.7])

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            roc_pivot('not explainer', protected, privileged)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            roc_pivot(exp, protected, privileged, cutoff=0.5, theta=1.4)

        with self.assertRaises(dx.fairness._basics.exceptions.ParameterCheckError):
            roc_pivot(exp, protected, privileged, cutoff=0.5, theta='sth_else')

if __name__ == '__main__':
    unittest.main()



from warnings import warn
import numpy as np


def yhat_default(m, d):
    return m.predict(d)


def yhat_proba_default(m, d):
    return m.predict_proba(d)[:, 1]


def yhat_xgboost(m, d):
    from xgboost import DMatrix
    return m.predict(DMatrix(d))


def get_tf_yhat(model):
    if not (str(type(model)).startswith("<class 'tensorflow.python.keras.engine") or
             str(type(model)).startswith("<class 'keras.engine")):
        return None

    if model.output_shape[1] == 1:
        return yhat_tf_regression, "regression"
    elif model.output_shape[1] == 2:
        return yhat_tf_classification, "classification"
    else:
        warn("Tensorflow: Output shape of the predict method should not be greater than 2")
        return yhat_tf_classification, "classification"


def yhat_tf_regression(m, d):
    return m.predict(np.array(d)).reshape(-1, )


def yhat_tf_classification(m, d):
    return m.predict(np.array(d))[:, 1]


def get_h2o_yhat(model):
    if not str(type(model)).startswith("<class 'h2o.estimators"):
        return None
    
    if model.type == 'classifier':
        return yhat_h2o_classification, "classification"
    if model.type == 'regressor':
        return yhat_h2o_regression, "regression"
    
    
def yhat_h2o_regression(m, d):
    from h2o import H2OFrame
    return m.predict(H2OFrame(d, column_types=m._column_types)).as_data_frame().to_numpy().flatten()


def yhat_h2o_classification(m, d):
    from h2o import H2OFrame
    return m.predict(H2OFrame(d, column_types=m._column_types)).as_data_frame().to_numpy()[:, 2]


# not used currently
def yhat_pycaret_regression(m, d):
    from pycaret.regression import predict_model
    return predict_model(m, d, verbose=False)['Score'].values
def yhat_pycaret_classification(m, d):
    from pycaret.classification import predict_model
    return predict_model(m, d, verbose=False)['Score'].values

    
def get_predict_function_and_model_type(model, model_class):
    
    prep_tf = get_tf_yhat(model)
    prep_h2o = get_h2o_yhat(model)
    
    # check for exceptions
    yhat_exception_dict = {
        "xgboost.core.Booster": (yhat_xgboost, None),  # there is no way of checking for regr/classif
        "tensorflow.python.keras.engine.sequential.Sequential": prep_tf,
        "tensorflow.python.keras.engine.training.Model": prep_tf,
        "tensorflow.python.keras.engine.functional.Functional": prep_tf,
        "keras.engine.sequential.Sequential": prep_tf,
        "keras.engine.training.Model": prep_tf,
        "keras.engine.functional.Functional": prep_tf,
        "h2o.estimators.coxph.H2OCoxProportionalHazardsEstimator": prep_h2o,
        "h2o.estimators.deeplearning.H2ODeepLearningEstimator": prep_h2o,
        "h2o.estimators.gam.H2OGeneralizedAdditiveEstimator": prep_h2o,
        "h2o.estimators.gbm.H2OGradientBoostingEstimator": prep_h2o,
        "h2o.estimators.glm.H2OGeneralizedLinearEstimator": prep_h2o,
        "h2o.estimators.naive_bayes.H2ONaiveBayesEstimator": prep_h2o,
        "h2o.estimators.psvm.H2OSupportVectorMachineEstimator": prep_h2o,
        "h2o.estimators.random_forest.H2ORandomForestEstimator": prep_h2o,
        "h2o.estimators.rulefit.H2ORuleFitEstimator": prep_h2o,
        "h2o.estimators.stackedensemble.H2OStackedEnsembleEstimator": prep_h2o,
        "h2o.estimators.targetencoder.H2OTargetEncoderEstimator": prep_h2o,
        "h2o.estimators.xgboost.H2OXGBoostEstimator": prep_h2o
    }

    if yhat_exception_dict.get(model_class, None) is not None:
        return yhat_exception_dict.get(model_class)

    model_type_ = None
    # additional extraction sklearn
    if hasattr(model, '_estimator_type'):
        if model._estimator_type == 'classifier':
            model_type_ = 'classification'
        elif model._estimator_type == 'regressor':
            model_type_ = 'regression'

    # default extraction
    if hasattr(model, 'predict_proba'):
        # check if model has predict_proba
        return yhat_proba_default, 'classification' if model_type_ is None else model_type_
    elif hasattr(model, 'predict'):
        # check if model has predict
        return yhat_default, 'regression' if model_type_ is None else model_type_
    else:
        # this path results in an error later
        return False, None



import numpy as np
import pandas as pd
from copy import deepcopy
import multiprocessing as mp

from dalex import _global_checks


def calculate_predict_aspect_importance(
    explainer,
    new_observation,
    variable_groups,
    N,
    n_aspects,
    sample_method,
    f,
    random_state
):
    _global_checks.global_check_import('scikit-learn', 'Predict Aspect Importance explanations')
    # create sample from data and copy
    r = np.random.RandomState(random_state)
    N = min(explainer.data.shape[0], N)
    ids = r.randint(0, explainer.data.shape[0], N)
    n_sample = explainer.data.iloc[ids, :].copy()
    n_sample_changed = deepcopy(n_sample)
    n_sample_changed.reset_index(inplace=True, drop=True)

    if len(variable_groups) == 1:
        result = pd.DataFrame(
            {
                "aspect_name": list(variable_groups.keys())[0],
                "variable_names": [list(variable_groups.values())[0]],
                "variable_values": new_observation[list(variable_groups.values())[0]].values.tolist(),
                "importance": explainer.predict(new_observation) - explainer.y_hat.mean(),
                "label": explainer.label,
            }
        )
        return result

    # generate binary matrix for replacing aspect
    X_prim = get_sample(N, len(variable_groups), sample_method, f, random_state)

    # replace aspect in sample with observation's values of features
    for i in range(n_sample_changed.shape[0]):
        for k, variables_set_key in enumerate(variable_groups):
            if X_prim[i, k] == 1:
                variables = variable_groups[variables_set_key]
                n_sample_changed.loc[i, variables] = [
                    new_observation.iloc[0][var] for var in variables
                ]

    # calculate change in predictions
    y_changed = explainer.predict(n_sample_changed) - explainer.predict(n_sample)

    # fit linear model to binary matrix and change in predictions vector to estimate aspect importance
    X_prim = pd.DataFrame(X_prim)
    X_prim.columns = [aspect_name for aspect_name in variable_groups]

    if n_aspects == None:
        from sklearn.linear_model import LinearRegression
        lr = LinearRegression()
        lr.fit(X_prim, y_changed)
        model_coef = lr.coef_
    else:
        from sklearn.linear_model import lasso_path
        lasso_res_matrix = lasso_path(X_prim, y_changed)[1]
        indx = np.max(np.where(np.count_nonzero(lasso_res_matrix, axis=0) <= n_aspects))
        model_coef = lasso_res_matrix[:, indx]

    # generate result pd.DataFrame
    variable_names = [
        variable_groups[variables_set_key] for variables_set_key in variable_groups
    ]
    variable_values = [
        new_observation[variables_list].values[0] for variables_list in variable_names
    ]
    result = pd.DataFrame(
        {
            "aspect_name": X_prim.columns,
            "variable_names": variable_names,
            "variable_values": variable_values,
            "importance": model_coef,
            "label": explainer.label,
        }
    )
    result = result.sort_values(by="importance", key=abs, ascending=False).reset_index(
        drop=True
    )
    return result


def calculate_shap_predict_aspect_importance(
         explainer,
         new_observation,
         variable_groups,
         N,
         B,
         processes,
         random_state
):
    r = np.random.RandomState(random_state)
    N = min(explainer.data.shape[0], N)
    ids = r.randint(0, explainer.data.shape[0], N)
    data_sample = explainer.data.iloc[ids, :].copy()
    p = len(variable_groups)

    if processes == 1:
        result_list = [
            iterate_paths(explainer.predict, data_sample,
                           new_observation, variable_groups, p, b + 1, np.random)
            for b in range(B)]
    else:
        # create number generator for each iteration
        ss = np.random.SeedSequence(random_state)
        generators = [np.random.default_rng(s) for s in ss.spawn(B)]
        pool = mp.get_context('spawn').Pool(processes)
        result_list = pool.starmap_async(iterate_paths,
                                         [(explainer.predict, data_sample,
                                           new_observation, variable_groups, p, b + 1, generators[b]) for b in range(B)]).get()
        pool.close()

    tmp_result = pd.concat(result_list)

    # average over all of the paths
    variable_average = tmp_result.pivot(index='aspect_name', columns='B', values='importance').mean(axis=1)
    
    # generate result pd.DataFrame
    result = pd.DataFrame(
            {
            "aspect_name": variable_average.index.tolist(), 
            "importance": variable_average.values.tolist(),
            "label": explainer.label
            }
        )
    result["variable_names"] = result["aspect_name"].map(variable_groups)
    variable_values = [
        new_observation[variables_list].values[0] for variables_list in result["variable_names"]
    ]
    result["variable_values"] = variable_values
    result = result.sort_values(by="importance", key=abs, ascending=False).reset_index(
        drop=True
    )
    return result[["aspect_name", "variable_names", "variable_values", "importance", "label"]]


def get_sample(n, p, sample_method="default", f=2, random_state=None):
    X_prim = np.zeros(shape=(n, p))
    def_rng = np.random.default_rng(random_state)
    # f is average number of replaced zeros for binomial sampling
    if sample_method == "binom":
        if f <= 0: 
            # it must be > 0 
            raise ValueError("f should be positive")
        if f >= p: 
            # it must be < p (number_of_aspects)
            f = p - 1
        for i in range(n):
            n_of_changes = max(def_rng.binomial(p, f / p, 1), 1)
            X_prim[i, def_rng.choice(p, n_of_changes, replace=False)] = 1
    else:
        for i in range(n):
            X_prim[i, np.unique(def_rng.integers(0, p, 2))] = 1
    return X_prim


def iterate_paths(predict_function, data_sample, new_observation, variable_groups, p, b, rng):
    random_path = rng.choice(np.arange(p), p, replace=False)
    return get_single_random_path(predict_function, data_sample, new_observation, variable_groups, random_path, b)


def get_single_random_path(predict_function, data_sample, new_observation, variable_groups, random_path, b):
    current_data = deepcopy(data_sample)
    current_data.reset_index(inplace=True, drop=True)
    variable_groups_varnames = list(variable_groups.values())
    variable_groups_aspnames = list(variable_groups.keys())
    yhats = np.empty(len(random_path) + 1)
    yhats[0] = predict_function(current_data).mean()
    for i, candidate in enumerate(random_path):
        variables = variable_groups_varnames[candidate]
        for var in variables:
            current_data.loc[:, var] = new_observation.iloc[0][var]
        yhats[i + 1] = predict_function(current_data).mean()

    diffs = np.diff(yhats)
    variable_names = [variable_groups_varnames[i] for i in random_path]
    aspect_name = [variable_groups_aspnames[i] for i in random_path]

    return pd.DataFrame({
        'aspect_name': aspect_name,
        'variable_names': variable_names,
        'importance': diffs,
        'B': b
    })



import unittest

import numpy as np
import pandas as pd
from plotly.graph_objs import Figure
from ipywidgets.widgets.widget_box import HBox
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn.ensemble import RandomForestRegressor
from traitlets.traitlets import Dict

import dalex as dx
from dalex.model_explanations._variable_importance.loss_functions import *
from dalex.aspect._predict_aspect_importance.object import PredictAspectImportance
from dalex.aspect._model_triplot.object import ModelTriplot
from dalex.aspect._predict_triplot.object import PredictTriplot
from dalex.aspect._model_aspect_importance.object import ModelAspectImportance

class AspectTestTitanic(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_titanic()
        data.loc[:, "survived"] = LabelEncoder().fit_transform(data.survived)

        self.X = data.drop(columns="survived")
        self.y = data.survived.values

        numeric_features = ["age", "fare", "sibsp", "parch"]
        numeric_transformer = Pipeline(
            steps=[
                ("imputer", SimpleImputer(strategy="median")),
                ("scaler", StandardScaler()),
            ]
        )

        categorical_features = ["gender", "class", "embarked"]
        categorical_transformer = Pipeline(
            steps=[
                ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
                ("onehot", OneHotEncoder(handle_unknown="ignore")),
            ]
        )

        preprocessor = ColumnTransformer(
            transformers=[
                ("num", numeric_transformer, numeric_features),
                ("cat", categorical_transformer, categorical_features),
            ]
        )

        clf = Pipeline(
            steps=[
                ("preprocessor", preprocessor),
                (
                    "classifier",
                    MLPClassifier(
                        hidden_layer_sizes=(50, 100, 50), max_iter=400, random_state=0
                    ),
                ),
            ]
        )

        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y, verbose=False)
        self.exp2 = dx.Explainer(clf, self.X, self.y, label="model2", verbose=False)
        self.aspect = dx.Aspect(self.exp)
        self.aspect2 = dx.Aspect(self.exp2, depend_method="pps")

    def test(self):
        self.assertIsInstance(self.aspect.explainer, dx.Explainer)
        self.assertIsInstance(self.aspect2.explainer, dx.Explainer)
        
        self.assertIsInstance(self.aspect.depend_matrix, pd.DataFrame)
        self.assertIsInstance(self.aspect2.depend_matrix, pd.DataFrame)
        pd.testing.assert_frame_equal(self.aspect.depend_matrix, self.aspect.depend_matrix.T)
        pd.testing.assert_frame_equal(self.aspect2.depend_matrix, self.aspect2.depend_matrix.T)
        for ind, row in self.aspect.depend_matrix.iterrows():
            self.assertEqual(row[ind], 1.0)
        for ind, row in self.aspect2.depend_matrix.iterrows():
            self.assertEqual(row[ind], 1.0)


        self.assertIsInstance(self.aspect.linkage_matrix, np.ndarray)
        self.assertIsInstance(self.aspect2.linkage_matrix, np.ndarray)

        self.assertIsInstance(self.aspect._hierarchical_clustering_dendrogram, Figure)
        self.assertIsInstance(self.aspect2._hierarchical_clustering_dendrogram, Figure)

        self.assertIsInstance(self.aspect._dendrogram_aspects_ordered, pd.DataFrame)
        self.assertIsInstance(self.aspect2._dendrogram_aspects_ordered, pd.DataFrame)
        
        self.assertIsInstance(self.aspect.get_aspects(h=0.3), dict)
        self.assertIsInstance(self.aspect2.get_aspects(h=0.99), dict)
        self.assertGreaterEqual(3, len(self.aspect.get_aspects(h=0.3, n=3)))
        self.assertEqual(len(self.aspect.depend_matrix), len(self.aspect2.get_aspects(h=3)))

        self.assertIsInstance(self.aspect.plot_dendrogram(show=False), Figure)
        self.assertIsInstance(self.aspect2.plot_dendrogram(show=False), Figure)
        
    def test_predict_parts(self):
        pai = self.aspect.predict_parts(self.X.iloc[12])
        groups = {"personal_info":["gender", "age"], "ticket_info":['class', 'embarked', 'fare'], 'other':['sibsp','parch']}
        pai2 = self.aspect2.predict_parts(self.X.iloc[12], variable_groups=groups)
        pai3 = self.aspect.predict_parts(self.X.iloc[19], type='shap')
        n_aspects = 4
        pai4 = self.aspect2.predict_parts(self.X.iloc[22], sample_method='binom', n_aspects=n_aspects)
        self.assertEqual(set(groups.keys()), set(pai2.result.aspect_name))

        for pai_x in [pai, pai2, pai3, pai4]:
            self.assertIsInstance(pai_x, dx.aspect.PredictAspectImportance)
            self.assertIsInstance(pai_x.result, pd.DataFrame)
            self.assertEqual(set(pai_x.result.columns), set(["aspect_name", "variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
            self.assertGreaterEqual(1, pai_x.result.min_depend.max())
            self.assertGreaterEqual(pai_x.result.min_depend.min(), 0)
            fig = pai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig2, Figure)
            fig3 = pai_x.plot(show=False, bar_width=23)
            self.assertIsInstance(fig3, Figure)
            fig4 = pai_x.plot(show=False, vertical_spacing=0.2)
            self.assertIsInstance(fig4, Figure)
            fig5 = pai_x.plot(show=False, max_aspects=3)
            self.assertIsInstance(fig5, Figure)
            fig6 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig6, Figure)
            for index, row in pai_x.result.iterrows():
                self.assertTrue(len(row['variable_values']) == len(row['variable_names']))
                self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))

        
    
    def test_predict_aspect_importance(self):
        pai = PredictAspectImportance(self.aspect.get_aspects(h=0.2))
        groups = {"personal_info":["gender", "age"], "ticket_info":['class', 'embarked', 'fare'], 'other':['sibsp','parch']}
        pai2 = PredictAspectImportance(variable_groups=groups)
        pai3 = PredictAspectImportance(self.aspect.get_aspects(h=0.5), type='shap')
        n_aspects = 4
        pai4 = PredictAspectImportance(self.aspect2.get_aspects(h=0.1), sample_method='binom', n_aspects=n_aspects)

        pai.fit(self.exp, self.X.iloc[12])
        pai2.fit(self.exp2, self.X.iloc[13])
        pai3.fit(self.exp, self.X.iloc[14])
        pai4.fit(self.exp2, self.X.iloc[15])

        self.assertEqual(set(groups.keys()), set(pai2.result.aspect_name))
        for pai_x in [pai, pai2, pai3, pai4]:
            self.assertIsInstance(pai_x, dx.aspect.PredictAspectImportance)
            self.assertIsInstance(pai_x.result, pd.DataFrame)
            self.assertEqual(set(pai_x.result.columns), set(["aspect_name", "variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
            self.assertGreaterEqual(1, pai_x.result.min_depend.max())
            self.assertGreaterEqual(pai_x.result.min_depend.min(), 0)
            fig = pai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig2, Figure)
            fig3 = pai_x.plot(show=False, bar_width=23)
            self.assertIsInstance(fig3, Figure)
            fig4 = pai_x.plot(show=False, vertical_spacing=0.2)
            self.assertIsInstance(fig4, Figure)
            fig5 = pai_x.plot(show=False, max_aspects=3)
            self.assertIsInstance(fig5, Figure)
            fig6 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig6, Figure)
            for index, row in pai_x.result.iterrows():
                self.assertTrue(len(row['variable_values']) == len(row['variable_names']))
                self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
        

    def test_model_parts(self):
        mai = self.aspect.model_parts()
        groups = {"personal_info":["gender", "age"], "ticket_info":['class', 'embarked', 'fare'], 'other':['sibsp','parch']}
        mai2 = self.aspect2.model_parts(variable_groups=groups)
        mai3 = self.aspect.model_parts(type='ratio')
        mai4 = self.aspect2.model_parts(type='difference')
        mai5 = self.aspect.model_parts(loss_function='rmse')

        self.assertEqual(mai.loss_function, loss_one_minus_auc)
        self.assertEqual(mai2.loss_function, loss_one_minus_auc)
        self.assertEqual(mai3.loss_function, loss_one_minus_auc)
        self.assertEqual(mai4.loss_function, loss_one_minus_auc)
        self.assertEqual(mai5.loss_function, loss_root_mean_square)

        for mai_x in [mai, mai2, mai3, mai4, mai5]:
            self.assertIsInstance(mai_x, dx.aspect.ModelAspectImportance)
            self.assertIsInstance(mai_x.result, pd.DataFrame)
            self.assertEqual(set(mai_x.result.columns), set(["aspect_name","variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
            self.assertGreaterEqual(1, mai_x.result.min_depend.max())
            self.assertGreaterEqual(mai_x.result.min_depend.min(), 0)
            for index, row in mai_x.result.iterrows():
                if row['aspect_name'] not in ['_baseline_', '_full_model_']:
                    self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
            fig = mai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = mai_x.plot(show=False, bar_width=18, vertical_spacing=0.12)
            self.assertIsInstance(fig2, Figure)
            fig3 = mai_x.plot(show=False, show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig3, Figure)
            fig4 = mai_x.plot(show=False,max_aspects=4,show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig4, Figure)

    def test_model_aspect_importance(self):
        mai = ModelAspectImportance(self.aspect.get_aspects(h=0.2))
        groups = {"personal_info":["gender", "age"], "ticket_info":['class', 'embarked', 'fare'], 'other':['sibsp','parch']}
        mai2 = ModelAspectImportance(variable_groups=groups)
        mai3 = ModelAspectImportance(self.aspect2.get_aspects(h=0.4), type='ratio')
        mai4 = ModelAspectImportance(self.aspect.get_aspects(h=0.3), type='difference')
        mai5 = ModelAspectImportance(self.aspect2.get_aspects(h=0.6), loss_function='rmse')

        mai.fit(self.exp)
        mai2.fit(self.exp)
        mai3.fit(self.exp2)
        mai4.fit(self.exp)
        mai5.fit(self.exp2)

        self.assertEqual(mai.loss_function, loss_one_minus_auc)
        self.assertEqual(mai2.loss_function, loss_one_minus_auc)
        self.assertEqual(mai3.loss_function, loss_one_minus_auc)
        self.assertEqual(mai4.loss_function, loss_one_minus_auc)
        self.assertEqual(mai5.loss_function, loss_root_mean_square)

        for mai_x in [mai, mai2, mai3, mai4, mai5]:
            self.assertIsInstance(mai_x, dx.aspect.ModelAspectImportance)
            self.assertIsInstance(mai_x.result, pd.DataFrame)
            self.assertEqual(set(mai_x.result.columns), set(["aspect_name","variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
            self.assertGreaterEqual(1, mai_x.result.min_depend.max())
            self.assertGreaterEqual(mai_x.result.min_depend.min(), 0)
            for index, row in mai_x.result.iterrows():
                if row['aspect_name'] not in ['_baseline_', '_full_model_']:
                    self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
            fig = mai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = mai_x.plot(show=False, bar_width=18, vertical_spacing=0.12)
            self.assertIsInstance(fig2, Figure)
            fig3 = mai_x.plot(show=False, show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig3, Figure)
            fig4 = mai_x.plot(show=False,max_aspects=4,show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig4, Figure)

    def test_predict_triplot(self):
        pt = self.aspect.predict_triplot(self.X.iloc[12], sample_method='binom')
        pt2 = self.aspect2.predict_triplot(self.X.iloc[12], type='shap')

        self.assertIsInstance(pt, dx.aspect.PredictTriplot)
        self.assertIsInstance(pt2, dx.aspect.PredictTriplot)

        self.assertIsInstance(pt.result, pd.DataFrame)
        self.assertIsInstance(pt2.result, pd.DataFrame)

        self.assertEqual(set(pt.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
        self.assertEqual(set(pt2.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))

        self.assertGreaterEqual(1, pt.result.min_depend.max())
        self.assertGreaterEqual(1, pt2.result.min_depend.max())

        self.assertGreaterEqual(pt.result.min_depend.min(), 0)
        self.assertGreaterEqual(pt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(pt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(pt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )

        for i in range(len(pt.result)):
            self.assertTrue(len(pt.result.loc[i, 'variable_values']) == len(pt.result.loc[i, 'variable_names']))
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        for i in range(len(pt2.result)):
            self.assertTrue([len(pt2.result.loc[i, 'variable_values']) == len(pt2.result.loc[i, 'variable_names'])])
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        
        fig = pt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = pt2.plot(show=False)
        self.assertIsInstance(fig2, Figure)
        fig3 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6)
        self.assertIsInstance(fig3, Figure)
        fig4 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6, widget=True)
        self.assertIsInstance(fig4, HBox)
        fig5 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test')
        self.assertIsInstance(fig5, Figure)
        fig6 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig6, HBox)
        fig7 = pt.plot(show=False, absolute_value=True, digits=2, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig7, HBox)

    
    def test_predict_triplot_class(self):
        pt = PredictTriplot(sample_method='binom')
        pt2 = PredictTriplot(type='shap')
        pt.fit(self.aspect, self.X.iloc[12])
        pt2.fit(self.aspect2, self.X.iloc[12])

        self.assertIsInstance(pt, dx.aspect.PredictTriplot)
        self.assertIsInstance(pt2, dx.aspect.PredictTriplot)

        self.assertIsInstance(pt.result, pd.DataFrame)
        self.assertIsInstance(pt2.result, pd.DataFrame)

        self.assertEqual(set(pt.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
        self.assertEqual(set(pt2.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))

        self.assertGreaterEqual(1, pt.result.min_depend.max())
        self.assertGreaterEqual(1, pt2.result.min_depend.max())

        self.assertGreaterEqual(pt.result.min_depend.min(), 0)
        self.assertGreaterEqual(pt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(pt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(pt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )

        for i in range(len(pt.result)):
            self.assertTrue(len(pt.result.loc[i, 'variable_values']) == len(pt.result.loc[i, 'variable_names']))
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        for i in range(len(pt2.result)):
            self.assertTrue([len(pt2.result.loc[i, 'variable_values']) == len(pt2.result.loc[i, 'variable_names'])])
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        
        fig = pt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = pt2.plot(show=False)
        self.assertIsInstance(fig2, Figure)
        fig3 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6)
        self.assertIsInstance(fig3, Figure)
        fig4 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6, widget=True)
        self.assertIsInstance(fig4, HBox)
        fig5 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test')
        self.assertIsInstance(fig5, Figure)
        fig6 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig6, HBox)
        fig7 = pt.plot(show=False, absolute_value=True, digits=2, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig7, HBox)


    def test_model_triplot(self):
        mt = self.aspect.model_triplot()
        mt2 = self.aspect.model_triplot(type="difference")

        self.assertEqual(mt.loss_function, loss_one_minus_auc)
        self.assertEqual(mt2.loss_function, loss_one_minus_auc)

        self.assertIsInstance(mt, dx.aspect.ModelTriplot)
        self.assertIsInstance(mt2, dx.aspect.ModelTriplot)

        self.assertIsInstance(mt.result, pd.DataFrame)
        self.assertIsInstance(mt2.result, pd.DataFrame)

        self.assertEqual(set(mt.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
        self.assertEqual(set(mt2.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))

        self.assertGreaterEqual(1, mt.result.min_depend.max())
        self.assertGreaterEqual(1, mt2.result.min_depend.max())

        self.assertGreaterEqual(mt.result.min_depend.min(), 0)
        self.assertGreaterEqual(mt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(mt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(mt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        
        fig = mt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = mt2.plot(show=False, width=1200)
        self.assertIsInstance(fig2, Figure)
        fig3 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test')
        self.assertIsInstance(fig3, Figure)
        fig4 = mt.plot(show=False, width=1200, show_change=False)
        self.assertIsInstance(fig4, Figure)
        fig5 = mt.plot(show=False, width=1200, show_change=False, widget=True)
        self.assertIsInstance(fig5, HBox)
        fig6 = mt.plot(show=False, digits=9, bar_width=22)
        self.assertIsInstance(fig6, Figure)
        fig7 = mt.plot(show=False, digits=9, bar_width=22, widget=True)
        self.assertIsInstance(fig7, HBox)
        fig8 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test', widget=True)
        self.assertIsInstance(fig8, HBox)

    
    def test_model_triplot_class(self):
        mt = ModelTriplot()
        mt2 = ModelTriplot(type="difference")

        mt.fit(self.aspect)
        mt2.fit(self.aspect2)

        self.assertEqual(mt.loss_function, loss_one_minus_auc)
        self.assertEqual(mt2.loss_function, loss_one_minus_auc)

        self.assertIsInstance(mt, dx.aspect.ModelTriplot)
        self.assertIsInstance(mt2, dx.aspect.ModelTriplot)

        self.assertIsInstance(mt.result, pd.DataFrame)
        self.assertIsInstance(mt2.result, pd.DataFrame)

        self.assertEqual(set(mt.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
        self.assertEqual(set(mt2.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))

        self.assertGreaterEqual(1, mt.result.min_depend.max())
        self.assertGreaterEqual(1, mt2.result.min_depend.max())

        self.assertGreaterEqual(mt.result.min_depend.min(), 0)
        self.assertGreaterEqual(mt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(mt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(mt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        
        fig = mt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = mt2.plot(show=False, width=1200)
        self.assertIsInstance(fig2, Figure)
        fig3 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test')
        self.assertIsInstance(fig3, Figure)
        fig4 = mt.plot(show=False, width=1200, show_change=False)
        self.assertIsInstance(fig4, Figure)
        fig5 = mt.plot(show=False, width=1200, show_change=False, widget=True)
        self.assertIsInstance(fig5, HBox)
        fig6 = mt.plot(show=False, digits=9, bar_width=22)
        self.assertIsInstance(fig6, Figure)
        fig7 = mt.plot(show=False, digits=9, bar_width=22, widget=True)
        self.assertIsInstance(fig7, HBox)
        fig8 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test', widget=True)
        self.assertIsInstance(fig8, HBox)


class AspectTestFifa(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_fifa()
        data = data.sample(int(len(data)*0.2))
        data = data.loc[:,['value_eur', 'age', 'attacking_short_passing', 'skill_dribbling', 'skill_ball_control', 'skill_long_passing', 'movement_acceleration', 'movement_reactions', 'mentality_vision', 'mentality_penalties','mentality_composure', 'defending_marking', 'defending_standing_tackle',
       'defending_sliding_tackle','power_jumping', 'power_stamina', 'power_strength']]

        self.X = data.drop(["value_eur"], axis = 1)
        self.y = data['value_eur']
        clf = Pipeline(steps=[
            ('classifier', RandomForestRegressor(random_state=123))
        ])

        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y, verbose=False)
        self.exp2 = dx.Explainer(clf, self.X, self.y, label="model2", verbose=False)
        self.aspect = dx.Aspect(self.exp)
        self.aspect2 = dx.Aspect(self.exp2, depend_method="pps")

    def test(self):
        self.assertIsInstance(self.aspect.explainer, dx.Explainer)
        self.assertIsInstance(self.aspect2.explainer, dx.Explainer)
        
        self.assertIsInstance(self.aspect.depend_matrix, pd.DataFrame)
        self.assertIsInstance(self.aspect2.depend_matrix, pd.DataFrame)
        pd.testing.assert_frame_equal(self.aspect2.depend_matrix, self.aspect2.depend_matrix.T)
        for ind, row in self.aspect.depend_matrix.iterrows():
            self.assertEqual(row[ind], 1.0)
        for ind, row in self.aspect2.depend_matrix.iterrows():
            self.assertEqual(row[ind], 1.0)


        self.assertIsInstance(self.aspect.linkage_matrix, np.ndarray)
        self.assertIsInstance(self.aspect2.linkage_matrix, np.ndarray)

        self.assertIsInstance(self.aspect._hierarchical_clustering_dendrogram, Figure)
        self.assertIsInstance(self.aspect2._hierarchical_clustering_dendrogram, Figure)

        self.assertIsInstance(self.aspect._dendrogram_aspects_ordered, pd.DataFrame)
        self.assertIsInstance(self.aspect2._dendrogram_aspects_ordered, pd.DataFrame)
        
        self.assertIsInstance(self.aspect.get_aspects(h=0.3), dict)
        self.assertIsInstance(self.aspect2.get_aspects(h=0.99), dict)
        self.assertGreaterEqual(3, len(self.aspect.get_aspects(h=0.3, n=3)))
        self.assertEqual(len(self.aspect.depend_matrix), len(self.aspect2.get_aspects(h=3)))

        self.assertIsInstance(self.aspect.plot_dendrogram(show=False), Figure)
        self.assertIsInstance(self.aspect2.plot_dendrogram(show=False), Figure)
        
    def test_predict_parts(self):
        pai = self.aspect.predict_parts(self.X.iloc[12])
        groups = self.aspect.get_aspects(h=0.3)
        pai2 = self.aspect2.predict_parts(self.X.iloc[12], variable_groups=groups)
        pai3 = self.aspect.predict_parts(self.X.iloc[19], type='shap')
        n_aspects = 4
        pai4 = self.aspect2.predict_parts(self.X.iloc[22], sample_method='binom', n_aspects=n_aspects)
        self.assertEqual(set(groups.keys()), set(pai2.result.aspect_name))

        for pai_x in [pai, pai2, pai3, pai4]:
            self.assertIsInstance(pai_x, dx.aspect.PredictAspectImportance)
            self.assertIsInstance(pai_x.result, pd.DataFrame)
            self.assertEqual(set(pai_x.result.columns), set(["aspect_name", "variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
            self.assertGreaterEqual(1, pai_x.result.min_depend.max())
            self.assertGreaterEqual(pai_x.result.min_depend.min(), 0)
            fig = pai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig2, Figure)
            fig3 = pai_x.plot(show=False, bar_width=23)
            self.assertIsInstance(fig3, Figure)
            fig4 = pai_x.plot(show=False, vertical_spacing=0.2)
            self.assertIsInstance(fig4, Figure)
            fig5 = pai_x.plot(show=False, max_aspects=3)
            self.assertIsInstance(fig5, Figure)
            fig6 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig6, Figure)
            for index, row in pai_x.result.iterrows():
                self.assertTrue(len(row['variable_values']) == len(row['variable_names']))
                self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
    
    def test_predict_aspect_importance(self):
        pai = PredictAspectImportance(self.aspect.get_aspects(h=0.2))
        groups = self.aspect.get_aspects(h=0.3)
        pai2 = PredictAspectImportance(variable_groups=groups)
        pai3 = PredictAspectImportance(self.aspect.get_aspects(h=0.5), type='shap')
        n_aspects = 4
        pai4 = PredictAspectImportance(self.aspect2.get_aspects(h=0.1), sample_method='binom', n_aspects=n_aspects)

        pai.fit(self.exp, self.X.iloc[12])
        pai2.fit(self.exp2, self.X.iloc[13])
        pai3.fit(self.exp, self.X.iloc[14])
        pai4.fit(self.exp2, self.X.iloc[15])

        self.assertEqual(set(groups.keys()), set(pai2.result.aspect_name))

        for pai_x in [pai, pai2, pai3, pai4]:
            self.assertIsInstance(pai_x, dx.aspect.PredictAspectImportance)
            self.assertIsInstance(pai_x.result, pd.DataFrame)
            self.assertEqual(set(pai_x.result.columns), set(["aspect_name", "variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
            self.assertGreaterEqual(1, pai_x.result.min_depend.max())
            self.assertGreaterEqual(pai_x.result.min_depend.min(), 0)
            fig = pai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig2, Figure)
            fig3 = pai_x.plot(show=False, bar_width=23)
            self.assertIsInstance(fig3, Figure)
            fig4 = pai_x.plot(show=False, vertical_spacing=0.2)
            self.assertIsInstance(fig4, Figure)
            fig5 = pai_x.plot(show=False, max_aspects=3)
            self.assertIsInstance(fig5, Figure)
            fig6 = pai_x.plot(show=False, max_aspects=3, digits=4, bar_width=23, title='Test plot', vertical_spacing=0.2)
            self.assertIsInstance(fig6, Figure)
            for index, row in pai_x.result.iterrows():
                self.assertTrue(len(row['variable_values']) == len(row['variable_names']))
                self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))

        

    def test_model_parts(self):
        mai = self.aspect.model_parts()
        groups = self.aspect.get_aspects(h=0.3)
        mai2 = self.aspect2.model_parts(variable_groups=groups)
        mai3 = self.aspect.model_parts(type='ratio')
        mai4 = self.aspect2.model_parts(type='difference')
        mai5 = self.aspect.model_parts(loss_function='1-auc')

        self.assertEqual(mai.loss_function, loss_root_mean_square)
        self.assertEqual(mai2.loss_function, loss_root_mean_square)
        self.assertEqual(mai3.loss_function, loss_root_mean_square)
        self.assertEqual(mai4.loss_function, loss_root_mean_square)
        self.assertEqual(mai5.loss_function, loss_one_minus_auc)

        for mai_x in [mai, mai2, mai3, mai4, mai5]:
            self.assertIsInstance(mai_x, dx.aspect.ModelAspectImportance)
            self.assertIsInstance(mai_x.result, pd.DataFrame)
            self.assertEqual(set(mai_x.result.columns), set(["aspect_name","variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
            self.assertGreaterEqual(1, mai_x.result.min_depend.max())
            self.assertGreaterEqual(mai_x.result.min_depend.min(), 0)
            for index, row in mai_x.result.iterrows():
                if row['aspect_name'] not in ['_baseline_', '_full_model_']:
                    self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
            fig = mai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = mai_x.plot(show=False, bar_width=18, vertical_spacing=0.12)
            self.assertIsInstance(fig2, Figure)
            fig3 = mai_x.plot(show=False, show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig3, Figure)
            fig4 = mai_x.plot(show=False,max_aspects=4,show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig4, Figure)
        

    def test_model_aspect_importance(self):
        mai = ModelAspectImportance(self.aspect.get_aspects(h=0.2))
        groups = self.aspect.get_aspects(h=0.3)
        mai2 = ModelAspectImportance(variable_groups=groups)
        mai3 = ModelAspectImportance(self.aspect2.get_aspects(h=0.4), type='ratio')
        mai4 = ModelAspectImportance(self.aspect.get_aspects(h=0.3), type='difference')
        mai5 = ModelAspectImportance(self.aspect2.get_aspects(h=0.6), loss_function='1-auc')

        mai.fit(self.exp)
        mai2.fit(self.exp)
        mai3.fit(self.exp2)
        mai4.fit(self.exp)
        mai5.fit(self.exp2)

        self.assertEqual(mai.loss_function, loss_root_mean_square)
        self.assertEqual(mai2.loss_function, loss_root_mean_square)
        self.assertEqual(mai3.loss_function, loss_root_mean_square)
        self.assertEqual(mai4.loss_function, loss_root_mean_square)
        self.assertEqual(mai5.loss_function, loss_one_minus_auc)

        self.assertEqual(set(groups.keys()), set(mai2.result.aspect_name[1:-1]))

        for mai_x in [mai, mai2, mai3, mai4, mai5]:
            self.assertIsInstance(mai_x, dx.aspect.ModelAspectImportance)
            self.assertIsInstance(mai_x.result, pd.DataFrame)
            self.assertEqual(set(mai_x.result.columns), set(["aspect_name","variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
            self.assertGreaterEqual(1, mai_x.result.min_depend.max())
            self.assertGreaterEqual(mai_x.result.min_depend.min(), 0)
            for index, row in mai_x.result.iterrows():
                if row['aspect_name'] not in ['_baseline_', '_full_model_']:
                    self.assertTrue(set(row['vars_min_depend']).issubset(set(row['variable_names'])))
            fig = mai_x.plot(show=False)
            self.assertIsInstance(fig, Figure)
            fig2 = mai_x.plot(show=False, bar_width=18, vertical_spacing=0.12)
            self.assertIsInstance(fig2, Figure)
            fig3 = mai_x.plot(show=False, show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig3, Figure)
            fig4 = mai_x.plot(show=False,max_aspects=4,show_variable_names=False, bar_width=18, title='Testing', vertical_spacing=0.12)
            self.assertIsInstance(fig4, Figure)

    def test_predict_triplot(self):
        pt = self.aspect.predict_triplot(self.X.iloc[12], sample_method='binom')
        pt2 = self.aspect2.predict_triplot(self.X.iloc[12], type='shap')

        self.assertIsInstance(pt, dx.aspect.PredictTriplot)
        self.assertIsInstance(pt2, dx.aspect.PredictTriplot)

        self.assertIsInstance(pt.result, pd.DataFrame)
        self.assertIsInstance(pt2.result, pd.DataFrame)

        self.assertEqual(set(pt.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
        self.assertEqual(set(pt2.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))

        self.assertGreaterEqual(1, pt.result.min_depend.max())
        self.assertGreaterEqual(1, pt2.result.min_depend.max())

        self.assertGreaterEqual(pt.result.min_depend.min(), 0)
        self.assertGreaterEqual(pt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(pt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(pt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )

        for i in range(len(pt.result)):
            self.assertTrue(len(pt.result.loc[i, 'variable_values']) == len(pt.result.loc[i, 'variable_names']))
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        for i in range(len(pt2.result)):
            self.assertTrue([len(pt2.result.loc[i, 'variable_values']) == len(pt2.result.loc[i, 'variable_names'])])
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        
        fig = pt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = pt2.plot(show=False)
        self.assertIsInstance(fig2, Figure)
        fig3 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6)
        self.assertIsInstance(fig3, Figure)
        fig4 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6, widget=True)
        self.assertIsInstance(fig4, HBox)
        fig5 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test')
        self.assertIsInstance(fig5, Figure)
        fig6 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig6, HBox)
        fig7 = pt.plot(show=False, absolute_value=True, digits=2, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig7, HBox)
    
    def test_predict_triplot_class(self):
        pt = PredictTriplot(sample_method='binom')
        pt2 = PredictTriplot(type='shap')
        pt.fit(self.aspect, self.X.iloc[12])
        pt2.fit(self.aspect2, self.X.iloc[12])

        self.assertIsInstance(pt, dx.aspect.PredictTriplot)
        self.assertIsInstance(pt2, dx.aspect.PredictTriplot)

        self.assertIsInstance(pt.result, pd.DataFrame)
        self.assertIsInstance(pt2.result, pd.DataFrame)

        self.assertEqual(set(pt.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))
        self.assertEqual(set(pt2.result.columns), set(["variable_names", "variable_values", "importance", "min_depend", "vars_min_depend", "label"]))

        self.assertGreaterEqual(1, pt.result.min_depend.max())
        self.assertGreaterEqual(1, pt2.result.min_depend.max())

        self.assertGreaterEqual(pt.result.min_depend.min(), 0)
        self.assertGreaterEqual(pt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(pt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(pt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )

        for i in range(len(pt.result)):
            self.assertTrue(len(pt.result.loc[i, 'variable_values']) == len(pt.result.loc[i, 'variable_names']))
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        for i in range(len(pt2.result)):
            self.assertTrue([len(pt2.result.loc[i, 'variable_values']) == len(pt2.result.loc[i, 'variable_names'])])
            self.assertTrue(set(pt.result.loc[i, 'vars_min_depend']).issubset(set(pt.result.loc[i, 'variable_names'])))
        
        fig = pt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = pt2.plot(show=False)
        self.assertIsInstance(fig2, Figure)
        fig3 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6)
        self.assertIsInstance(fig3, Figure)
        fig4 = pt.plot(show=False, absolute_value=True, width=1200, abbrev_labels=6, widget=True)
        self.assertIsInstance(fig4, HBox)
        fig5 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test')
        self.assertIsInstance(fig5, Figure)
        fig6 = pt.plot(show=False, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig6, HBox)
        fig7 = pt.plot(show=False, absolute_value=True, digits=2, bar_width=22, width=1200, abbrev_labels=6, title='Test', widget=True)
        self.assertIsInstance(fig7, HBox)


    def test_model_triplot(self):
        mt = self.aspect.model_triplot()
        mt2 = self.aspect.model_triplot(type="difference")

        self.assertEqual(mt.loss_function, loss_root_mean_square)
        self.assertEqual(mt2.loss_function, loss_root_mean_square)

        self.assertIsInstance(mt, dx.aspect.ModelTriplot)
        self.assertIsInstance(mt2, dx.aspect.ModelTriplot)

        self.assertIsInstance(mt.result, pd.DataFrame)
        self.assertIsInstance(mt2.result, pd.DataFrame)

        self.assertEqual(set(mt.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
        self.assertEqual(set(mt2.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))

        self.assertGreaterEqual(1, mt.result.min_depend.max())
        self.assertGreaterEqual(1, mt2.result.min_depend.max())

        self.assertGreaterEqual(mt.result.min_depend.min(), 0)
        self.assertGreaterEqual(mt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(mt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(mt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        
        fig = mt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = mt2.plot(show=False, width=1200)
        self.assertIsInstance(fig2, Figure)
        fig3 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test')
        self.assertIsInstance(fig3, Figure)
        fig4 = mt.plot(show=False, width=1200, show_change=False)
        self.assertIsInstance(fig4, Figure)
        fig5 = mt.plot(show=False, width=1200, show_change=False, widget=True)
        self.assertIsInstance(fig5, HBox)
        fig6 = mt.plot(show=False, digits=9, bar_width=22)
        self.assertIsInstance(fig6, Figure)
        fig7 = mt.plot(show=False, digits=9, bar_width=22, widget=True)
        self.assertIsInstance(fig7, HBox)
        fig8 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test', widget=True)
        self.assertIsInstance(fig8, HBox)
    
    def test_model_triplot_class(self):
        mt = ModelTriplot()
        mt2 = ModelTriplot(type="difference")

        mt.fit(self.aspect)
        mt2.fit(self.aspect2)

        self.assertEqual(mt.loss_function, loss_root_mean_square)
        self.assertEqual(mt2.loss_function, loss_root_mean_square)

        self.assertIsInstance(mt, dx.aspect.ModelTriplot)
        self.assertIsInstance(mt2, dx.aspect.ModelTriplot)

        self.assertIsInstance(mt.result, pd.DataFrame)
        self.assertIsInstance(mt2.result, pd.DataFrame)

        self.assertEqual(set(mt.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))
        self.assertEqual(set(mt2.result.columns), set(["variable_names","dropout_loss","dropout_loss_change","label", "min_depend", "vars_min_depend"]))

        self.assertGreaterEqual(1, mt.result.min_depend.max())
        self.assertGreaterEqual(1, mt2.result.min_depend.max())

        self.assertGreaterEqual(mt.result.min_depend.min(), 0)
        self.assertGreaterEqual(mt2.result.min_depend.min(), 0)

        self.assertEqual(
            set(mt.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        self.assertEqual(
            set(mt2.result["variable_names"].iloc[-1]),
            set(self.X.columns),
        )
        
        fig = mt.plot(show=False)
        self.assertIsInstance(fig, Figure)
        fig2 = mt2.plot(show=False, width=1200)
        self.assertIsInstance(fig2, Figure)
        fig3 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test')
        self.assertIsInstance(fig3, Figure)
        fig4 = mt.plot(show=False, width=1200, show_change=False)
        self.assertIsInstance(fig4, Figure)
        fig5 = mt.plot(show=False, width=1200, show_change=False, widget=True)
        self.assertIsInstance(fig5, HBox)
        fig6 = mt.plot(show=False, digits=9, bar_width=22)
        self.assertIsInstance(fig6, Figure)
        fig7 = mt.plot(show=False, digits=9, bar_width=22, widget=True)
        self.assertIsInstance(fig7, HBox)
        fig8 = mt.plot(show=False, digits=2, bar_width=22, width=1200, title='Test', widget=True)
        self.assertIsInstance(fig8, HBox)
        



# helper functions for checks
import re
import numpy as np


def verbose_cat(text, verbose):
    if verbose:
        print(text)


def is_y_in_data(data, y):
    return (np.apply_along_axis(lambda x, y: (x == y).all(), 0, data, y)).any()


def get_model_info(model):
    model_package = re.search("(?<=<class ').*?(?=\.)", str(type(model)))[0]
    return {'model_package': model_package}



import numpy as np
import pandas as pd
import plotly.express as px
import warnings
from copy import deepcopy

from . import checks, plot, utils
from ... import _theme, _global_checks, _global_utils
from ..._explanation import Explanation


class CeterisParibus(Explanation):
    """Calculate predict-level variable profiles as Ceteris Paribus

    Parameters
    -----------
    variables : array_like of str, optional
        Variables for which the profiles will be calculated
        (default is `None`, which means all of the variables).
    grid_points : int, optional
        Maximum number of points for profile calculations (default is `101`).
        NOTE: The final number of points may be lower than `grid_points`,
        eg. if there is not enough unique values for a given variable.
    variable_splits : dict of lists, optional
        Split points for variables e.g. `{'x': [0, 0.2, 0.5, 0.8, 1], 'y': ['a', 'b']}`
        (default is `None`, which means that they will be calculated using one of
        `variable_splits_type` and the `data` attribute).
    variable_splits_type : {'uniform', 'quantiles'}, optional
        Way of calculating `variable_splits`. Set `'quantiles'` for percentiles.
        (default is `'uniform'`, which means uniform grid of points).
    variable_splits_with_obs: bool, optional
        Add variable values of `new_observation` data to the `variable_splits`
        (default is `True`).
    processes : int, optional
        Number of parallel processes to use in calculations. Iterated over `variables`
        (default is `1`, which means no parallel computation).

    Attributes
    -----------
    result : pd.DataFrame
        Main result attribute of an explanation.
    new_observation : pd.DataFrame
        Observations for which predictions need to be explained.
    variables : array_like of str or None
        Variables for which the profiles will be calculated.
    grid_points : int
        Maximum number of points for profile calculations.
    variable_splits : dict of lists or None
        Split points for variables.
    variable_splits_type : {'uniform', 'quantiles'}
        Way of calculating `variable_splits`.
    variable_splits_with_obs: bool
        Add variable values of `new_observation` data to the `variable_splits`.
    processes : int
        Number of parallel processes to use in calculations. Iterated over `B`.

    Notes
    --------
    - https://pbiecek.github.io/ema/ceterisParibus.html
    """

    def __init__(self,
                 variables=None,
                 grid_points=101,
                 variable_splits=None,
                 variable_splits_type='uniform',
                 variable_splits_with_obs=False,
                 processes=1):

        _processes = checks.check_processes(processes)
        _variable_splits_type = checks.check_variable_splits_type(variable_splits_type)

        self.variables = variables
        self.grid_points = grid_points
        self.variable_splits = variable_splits
        self.variable_splits_type = _variable_splits_type
        self.variable_splits_with_obs = variable_splits_with_obs
        self.result = pd.DataFrame()
        self.new_observation = None
        self.processes = _processes

    def _repr_html_(self):
        return self.result._repr_html_()

    def fit(self,
            explainer,
            new_observation,
            y=None,
            verbose=True):
        """Calculate the result of explanation

        Fit method makes calculations in place and changes the attributes.

        Parameters
        -----------
        explainer : Explainer object
            Model wrapper created using the Explainer class.
        new_observation : pd.DataFrame or np.ndarray
            Observations for which predictions need to be explained.
        y : pd.Series or np.ndarray (1d), optional
            Target variable with the same length as `new_observation`.
        verbose : bool, optional
            Print tqdm progress bar (default is `True`).

        Returns
        -----------
        None
        """

        self.variables = checks.check_variables(self.variables, explainer, self.variable_splits)

        checks.check_data(explainer.data, self.variables)

        self.new_observation = checks.check_new_observation(new_observation, explainer)

        self.variable_splits = checks.check_variable_splits(self.variable_splits,
                                                     self.variables,
                                                     self.grid_points,
                                                     explainer.data,
                                                     self.variable_splits_type,
                                                     self.variable_splits_with_obs,
                                                     self.new_observation)

        y = checks.check_y(y)

        self.result, self.new_observation = utils.calculate_ceteris_paribus(
            explainer,
            self.new_observation,
            self.variable_splits,
            y,
            self.processes,
            verbose
        )

    def plot(self,
             objects=None,
             variable_type="numerical",
             variables=None,
             size=2,
             alpha=1,
             color="_label_",
             facet_ncol=2,
             facet_scales="free_x",
             show_observations=True,
             title="Ceteris Paribus Profiles",
             y_title='prediction',
             horizontal_spacing=None,
             vertical_spacing=None,
             show=True):
        """Plot the Ceteris Paribus explanation

        Parameters
        -----------
        objects : CeterisParibus object or array_like of CeterisParibus objects
            Additional objects to plot in subplots (default is `None`).
        variable_type : {'numerical', 'categorical'}
            Plot the profiles for numerical or categorical variables 
            (default is `'numerical'`).
        variables : str or array_like of str, optional
            Variables for which the profiles will be calculated
            (default is `None`, which means all of the variables).
        size : float, optional
            Width of lines in px (default is `2`).
        alpha : float <0, 1>, optional
            Opacity of lines (default is `1`).
        color : str, optional
            Variable name used for grouping
            (default is `'_label_'`, which groups by models).
        facet_ncol : int, optional
            Number of columns on the plot grid (default is `2`).
        facet_scales : {'free_x', 'free'}, optional
            Set to `free` for free y-axis range (default is `'free_x'`).
        show_observations : bool, optional
            Show observation points (default is `True`).
        title : str, optional
            Title of the plot (default is `"Ceteris Paribus Profiles"`).
        y_title : str, optional
            Title of the y/x axis (default is `"prediction"`).
        horizontal_spacing : float <0, 1>, optional
            Ratio of horizontal space between the plots
            (default depends on `variable_type`).
        vertical_spacing : float <0, 1>, optional
            Ratio of vertical space between the plots (default is `0.3/number of rows`).
        show : bool, optional
            `True` shows the plot; `False` returns the plotly Figure object that can 
            be edited or saved using the `write_image()` method (default is `True`).

        Returns
        -----------
        None or plotly.graph_objects.Figure
            Return figure that can be edited or saved. See `show` parameter.
        """

        if variable_type not in ("numerical", "categorical"):
            raise TypeError("variable_type should be 'numerical' or 'categorical'")
        if isinstance(variables, str):
            variables = (variables,)

        # are there any other objects to plot?
        if objects is None:
            _result_df = self.result.assign(_original_yhat_=lambda x: self.new_observation.loc[x.index, '_yhat_'])
            _include = self.variable_splits_with_obs
        elif isinstance(objects, self.__class__):  # allow for objects to be a single element
            _result_df = pd.concat([
                self.result.assign(_original_yhat_=lambda x: self.new_observation.loc[x.index, '_yhat_']),
                objects.result.assign(_original_yhat_=lambda x: objects.new_observation.loc[x.index, '_yhat_'])])
            _include = np.all([self.variable_splits_with_obs, objects.variable_splits_with_obs])
        elif isinstance(objects, (list, tuple)):  # objects as tuple or array
            _result_df = self.result.assign(_original_yhat_=lambda x: self.new_observation.loc[x.index, '_yhat_'])
            _include = [self.variable_splits_with_obs]
            for ob in objects:
                _global_checks.global_check_object_class(ob, self.__class__)
                _result_df = pd.concat([
                    _result_df, ob.result.assign(_original_yhat_=lambda x: ob.new_observation.loc[x.index, '_yhat_'])])
                _include += [ob.variable_splits_with_obs]
            _include = np.all(_include)
        else:
            _global_checks.global_raise_objects_class(objects, self.__class__)

        if _include is False and show_observations:
                warnings.warn("'show_observations' parameter changed to False,"
                              "because the 'variable_splits_with_obs' attribute is False"
                              "See `variable_splits_with_obs` parameter in `predict_profile`.")
                show_observations = False

        # variables to use
        all_variables = list(_result_df['_vname_'].dropna().unique())

        if variables is not None:
            all_variables = _global_utils.intersect_unsorted(variables, all_variables)
            if len(all_variables) == 0:
                raise TypeError("variables do not overlap with " + ''.join(variables))

        # names of numeric variables
        numeric_variables = _result_df[all_variables].select_dtypes(include=np.number).columns.tolist()

        if variable_type == "numerical":
            variable_names = numeric_variables

            if len(variable_names) == 0:
                # change to categorical
                variable_type = "categorical"
                # send message
                warnings.warn("'variable_type' parameter changed to 'categorical' due to lack of numerical variables.")
                # take all
                variable_names = all_variables
            elif variables is not None and len(variable_names) != len(variables):
                raise TypeError("There are no numerical variables")
        else:
            variable_names = np.setdiff1d(all_variables, numeric_variables).tolist()

            # there are variables selected
            if variables is not None:
                # take all
                variable_names = all_variables
            elif len(variable_names) == 0:
                # there were no variables selected and there are no categorical variables
                raise TypeError("There are no non-numerical variables.")

        # prepare profiles data
        _result_df = _result_df.loc[_result_df['_vname_'].isin(variable_names), ].reset_index(drop=True)

        #  calculate y axis range to allow for fixedrange True
        dl = _result_df['_yhat_'].to_numpy()
        min_max_margin = dl.ptp() * 0.10
        min_max = [dl.min() - min_max_margin, dl.max() + min_max_margin]

        # create _x_
        if len(variable_names) == 1:
            _result_df.loc[:, '_x_'] = deepcopy(_result_df.loc[:, variable_names[0]])
        else:
            for variable in variable_names:
                where_variable = _result_df['_vname_'] == variable
                _result_df.loc[where_variable, '_x_'] = deepcopy(_result_df.loc[where_variable, variable])

        # change x column to proper character values
        if variable_type == 'categorical':
            _result_df.loc[:, '_x_'] = _result_df.apply(lambda row: str(row[row['_vname_']]), axis=1)

        n = len(variable_names)
        facet_nrow = int(np.ceil(n / facet_ncol))
        if vertical_spacing is None:
            vertical_spacing = 0.3 / facet_nrow if variable_type == 'numerical' else 0.05
        if horizontal_spacing is None:
            horizontal_spacing = 0.05 if variable_type == 'numerical' else 0.1

        plot_height = 78 + 71 + facet_nrow * (280 + 60)

        _result_df = _result_df.assign(_text_=_result_df.apply(lambda obs: plot.tooltip_text(obs), axis=1))

        if variable_type == "numerical":    
            m = len(_result_df[color].dropna().unique())
            _result_df[color] = _result_df[color].astype(object)  # prevent error when using pd.StringDtype
        
            fig = px.line(_result_df,
                          x="_x_", y="_yhat_", color=color, facet_col="_vname_", line_group='_ids_',
                          category_orders={"_vname_": list(variable_names)},
                          labels={'_yhat_': 'prediction', '_label_': 'label', '_ids_': 'id'},  # , color: 'group'},
                          # hover_data={'_text_': True, '_yhat_': ':.3f', '_vname_': False, '_x_': False, color: False},
                          custom_data=['_text_'],
                          facet_col_wrap=facet_ncol,
                          facet_row_spacing=vertical_spacing,
                          facet_col_spacing=horizontal_spacing,
                          template="none",
                          color_discrete_sequence=_theme.get_default_colors(m, 'line')) \
                    .update_traces(dict(line_width=size, opacity=alpha,
                                        hovertemplate="%{customdata[0]}<extra></extra>")) \
                    .update_xaxes({'matches': None, 'showticklabels': True,
                                   'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True,
                                   'ticks': "outside", 'tickcolor': 'white', 'ticklen': 3, 'fixedrange': True}) \
                    .update_yaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True,
                                   'ticks': 'outside', 'tickcolor': 'white', 'ticklen': 3, 'fixedrange': True,
                                   'range': min_max})

            if show_observations:
                _points_df = _result_df.loc[_result_df['_original_'] == _result_df['_x_'], :].copy()

                fig_points = px.scatter(_points_df,
                                        x='_original_', y='_yhat_', facet_col='_vname_',
                                        category_orders={"_vname_": list(variable_names)},
                                        labels={'_yhat_': 'prediction', '_label_': 'label', '_ids_': 'id'},
                                        custom_data=['_text_'],
                                        facet_col_wrap=facet_ncol,
                                        facet_row_spacing=vertical_spacing,
                                        facet_col_spacing=horizontal_spacing,
                                        color_discrete_sequence=["#371ea3"]) \
                               .update_traces(dict(marker_size=5*size, opacity=alpha),
                                              hovertemplate="%{customdata[0]}<extra></extra>")

                for _, value in enumerate(fig_points.data):
                    fig.add_trace(value)
                    
            fig = _theme.fig_update_line_plot(fig, title, y_title, plot_height, 'closest')

        else:
            if color=="_label_" and len(_result_df['_ids_'].unique()) > 1 and len(_result_df['_label_'].unique()) == 1:
                warnings.warn("'color' parameter changed to '_ids_' because there are multiple observations for one model.")
                color = '_ids_'
            elif color=="_label_" and len(_result_df['_ids_'].unique()) > len(_result_df['_label_'].unique()): 
                # https://github.com/plotly/plotly.py/issues/2657
                raise TypeError("Please pick one observation per label or change the `color` parameter.")

            m = len(_result_df[color].dropna().unique())
            _result_df[color] = _result_df[color].astype(object)  # prevent error when using pd.StringDtype
            
            _result_df = _result_df.assign(_diff_=lambda x: x['_yhat_'] - x['_original_yhat_'])
            fig = px.bar(_result_df,
                         x="_diff_", y="_x_", color=color, facet_col="_vname_",
                         category_orders={"_vname_": list(variable_names)},
                         labels={'_yhat_': 'prediction', '_label_': 'label', '_ids_': 'id'},  # , color: 'group'},
                         # hover_data={'_yhat_': ':.3f', '_ids_': True, '_vname_': False, color: False},
                         custom_data=['_text_'],
                         base="_original_yhat_",
                         facet_col_wrap=facet_ncol,
                         facet_row_spacing=vertical_spacing,
                         facet_col_spacing=horizontal_spacing,
                         template="none",
                         color_discrete_sequence=_theme.get_default_colors(m, 'line'),
                         barmode='group',
                         orientation='h')  \
                    .update_traces(dict(opacity=alpha),
                                   hovertemplate="%{customdata[0]}<extra></extra>") \
                    .update_yaxes({'matches': None, 'showticklabels': True,
                                   'type': 'category', 'gridwidth': 2, 'automargin': True, 
                                   'ticks': "outside", 'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True}) \
                    .update_xaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True,
                                   'ticks': 'outside', 'tickcolor': 'white', 'ticklen': 3, 'fixedrange': True,
                                   'range': min_max})
                    
            for _, bar in enumerate(fig.data):
                fig.add_vline(x=bar.base[0], layer='below',
                              line={'color': "#371ea3", 'width': 1.5, 'dash': 'dot'})
                
            if show_observations:
                _points_df = _result_df.loc[_result_df['_original_'] == _result_df['_x_'], :].copy()

                fig_points = px.scatter(_points_df,
                                        x='_yhat_', y='_x_', facet_col='_vname_',
                                        category_orders={"_vname_": list(variable_names)},
                                        labels={'_yhat_': 'prediction', '_label_': 'label', '_ids_': 'id'},
                                        custom_data=['_text_'],
                                        facet_col_wrap=facet_ncol,
                                        facet_row_spacing=vertical_spacing,
                                        facet_col_spacing=horizontal_spacing,
                                        color_discrete_sequence=["#371ea3"]) \
                               .update_traces(dict(marker_size=5*size, opacity=alpha),
                                              hovertemplate="%{customdata[0]}<extra></extra>")

                for _, value in enumerate(fig_points.data):
                    fig.add_trace(value)

            fig = _theme.fig_update_bar_plot(fig, title, y_title, plot_height, 'closest')
            
        fig.update_layout(hoverlabel=dict(bgcolor='rgba(0,0,0,0.8)'))

        if facet_scales == "free":
            fig.update_yaxes(matches=None, range=None, showticklabels=True)

        if show:
            fig.show(config=_theme.get_default_config())
        else:
            return fig



import unittest

from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.neural_network import MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn.ensemble import RandomForestClassifier
from dalex.wrappers import ShapWrapper

import dalex as dx


class TestShapWrapperMLPRegressorTitanicFullDataset(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_titanic()
        data.loc[:, 'survived'] = LabelEncoder().fit_transform(data.survived)

        self.X = data.drop(columns='survived')
        self.y = data.survived

        numeric_features = ['age', 'fare', 'sibsp', 'parch']
        numeric_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='median')),
            ('scaler', StandardScaler())])

        categorical_features = ['gender', 'class', 'embarked']
        categorical_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='constant', fill_value='missing')),
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)])

        clf = Pipeline(steps=[('preprocessor', preprocessor),
                              ('classifier', MLPRegressor(hidden_layer_sizes=(150, 100, 50),
                                                          max_iter=500, random_state=0))])

        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y, verbose=False)

    def test(self):
        with self.assertRaises(TypeError):
            self.exp.predict_parts(self.X.iloc[[0]], type='shap_wrapper')


class TestShapWrapperRandomForestClassifierTitanicNumericalDataset(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_titanic()
        data.loc[:, 'survived'] = LabelEncoder().fit_transform(data.survived)

        self.X = data.loc[:, ["age", "fare", "sibsp", "parch"]]
        self.y = data.survived

        clf = RandomForestClassifier(n_estimators=100, random_state=123)
        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y, verbose=False)

    def test_predict_parts(self):
        case1 = self.exp.predict_parts(self.X.iloc[[0]], type='shap_wrapper')
        case2 = self.exp.predict_parts(self.X.iloc[1:2, :], type='shap_wrapper', shap_explainer_type='KernelExplainer')
        case3 = self.exp.predict_parts(self.X.iloc[1, :], type='shap_wrapper')
        case4 = self.exp.predict_parts(self.X.iloc[[0]], N=500, type='shap_wrapper', shap_explainer_type='KernelExplainer')

        self.assertIsInstance(case1, ShapWrapper)
        self.assertEqual(case1.shap_explainer_type, 'TreeExplainer')

        self.assertIsInstance(case2, ShapWrapper)
        self.assertEqual(case2.shap_explainer_type, 'KernelExplainer')

        self.assertIsInstance(case3, ShapWrapper)
        self.assertIsInstance(case4, ShapWrapper)

        case1.plot(show=False)
        case2.plot(show=False)
        case3.plot(show=False)
        case4.plot(show=False)

    def test_model_parts(self):
        case1 = self.exp.model_parts(type='shap_wrapper', N=22)
        case2 = self.exp.model_parts(type='shap_wrapper', N=22, shap_explainer_type='KernelExplainer')

        self.assertIsInstance(case1, ShapWrapper)
        self.assertEqual(case1.shap_explainer_type, 'TreeExplainer')

        self.assertIsInstance(case2, ShapWrapper)
        self.assertEqual(case2.shap_explainer_type, "KernelExplainer")

        case1.plot(show=False)
        case2.plot(show=False)


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from warnings import warn

import numpy as np
import pandas as pd


def check_columns_in_new_observation(new_observation,
                                     explainer):
    if not set(new_observation.columns).issubset(explainer.data):
        raise ValueError("Columns in new observation does not match these in training dataset.")


def check_new_observation(new_observation, explainer):
    new_observation_ = deepcopy(new_observation)
    if isinstance(new_observation_, pd.Series):
        new_observation_ = new_observation_.to_frame().T
        new_observation_.columns = explainer.data.columns
    elif isinstance(new_observation_, np.ndarray):
        if new_observation_.ndim == 1:
            # make 1D array 2D
            new_observation_ = new_observation_.reshape((1, -1))
        elif new_observation_.ndim > 2:
            raise ValueError("Wrong new_observation dimension")
        elif new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_ = pd.DataFrame(new_observation_)
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, list):
        new_observation_ = pd.DataFrame(new_observation_).T
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, pd.DataFrame):
        if new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_.columns = explainer.data.columns
    else:
        raise TypeError("new_observation must be a numpy.ndarray or pandas.Series or pandas.DataFrame")

    if pd.api.types.is_bool_dtype(new_observation_.index):
        raise ValueError("new_observation index is of boolean type")

    return new_observation_


def check_path(path):
    if not isinstance(path, (str, list, np.ndarray, pd.Series)):
        raise TypeError('path must be str or list or numpy.ndarray or pd.Series')
    if isinstance(path, str):
        if path != 'average':
            raise ValueError("if path is str, then it must be 'average'")
        else:
            return path

    path_ = np.array(path)
    if not np.issubdtype(path_[0], int):
        raise TypeError('if path is array-like, then must be array of integers')

    return path_


def check_processes(processes):
    from multiprocessing import cpu_count
    if processes > cpu_count():
        warn("You have asked for too many processes. Truncated to the number of physical CPUs.")

        return cpu_count()

    else:
        return processes


def check_random_state(random_state):
    if random_state is not None:
        np.random.seed(random_state)

    return random_state
